import pytest
from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_sdk.events.log_events import LogEvent
from acceldata_sdk.models.job import JobMetadata, Node
from acceldata_sdk.models.pipeline import CreatePipeline, PipelineMetadata
from acceldata_sdk.torch_client import TorchClient
from acceldata_sdk.models.job import CreateJob
from acceldata_sdk.models.pipeline import PipelineRunResult, PipelineRunStatus
from datetime import datetime, timedelta
import test_constants as test_const

# Importing the pprint module for pretty printing the output
import pprint

pp = pprint.PrettyPrinter(indent=4)

# UID of the pipeline in torch
pipeline_uid = "torch.external.integration.demo"
# Name of the pipeline in torch
pipeline_name = "Flow ETL External Integration Demo"
# Name of the postgres datasource in torch
postgres_ds = "POSTGRES-DS"
# Name of the s3 datasource in torch
s3_ds = "S3-DS"
# Name of initial output asset which is the output of job copying generated csv data to s3 as csv
s3_customer = "s3_customers"
# Name of output asset which is the output of job joining data from 2nd csv and postgres by looking up department
# name for department id
s3_postgres_customers = "s3_customers_postgres"
# Name of postgres table where department id vs department name mapping has been created in demo schema
pg_table = "demo.dept_info"

pg_database = "customers"
# Name of postgres field to order sql query on
order_by = "customer_id"
# S3 bucket name
bucket = "airflow-sdk-demo"
# The location of initial csv file in s3 bucket
csv1 = 'demo/customers.csv'
# The location of csv file in s3 bucket after joining with data from postgres table
csv3 = 'demo/customers-with-department-info.csv'
# Number of entries to be generated
csv_size = 30

# Global variables needed across functions
parent_span_context = None
pipeline_run_id = None
currentime = datetime.now()
explicit_pipeline_createdAt = currentime - timedelta(days=2, hours=6, minutes=30)
explicit_pipeline_updatedAt = explicit_pipeline_createdAt + timedelta(minutes=30)
explicit_pipeline_run_startedAt = explicit_pipeline_createdAt + timedelta(minutes=30)
root_span_created_at = explicit_pipeline_run_startedAt + timedelta(minutes=1)

child_span_1_created_at = explicit_pipeline_run_startedAt + timedelta(minutes=30)
child_span_1_finished_at = child_span_1_created_at + timedelta(minutes=30)

child_span_2_created_at = child_span_1_finished_at + timedelta(minutes=1)
child_span_2_finished_at = child_span_2_created_at + timedelta(minutes=10)

child_span_3_created_at = child_span_2_finished_at + timedelta(minutes=1)

child_span_3_child_1_created_at = child_span_3_created_at + timedelta(minutes=1)
child_span_3_child_1_finished_at = child_span_3_child_1_created_at + timedelta(minutes=10)

child_span_3_child_2_created_at = child_span_3_child_1_finished_at + timedelta(minutes=1)

child_span_3_child_2_child_1_created_at = child_span_3_child_2_created_at + timedelta(minutes=1)
child_span_3_child_2_child_1_finished_at = child_span_3_child_2_child_1_created_at + timedelta(minutes=10)

child_span_3_child_2_child_2_created_at = child_span_3_child_2_child_1_finished_at + timedelta(minutes=1)
child_span_3_child_2_child_2_finished_at = child_span_3_child_2_child_2_created_at + timedelta(minutes=10)

child_span_3_child_2_finished_at = child_span_3_child_2_child_2_finished_at
child_span_3_child_3_created_at = child_span_3_child_2_finished_at + timedelta(minutes=1)
child_span_3_child_3_finished_at = child_span_3_child_3_created_at + timedelta(minutes=10)
child_span_3_finished_at = child_span_3_child_3_finished_at
root_span_finishedAt = child_span_3_finished_at + timedelta(minutes=1)
explicit_pipeline_run_finishedAt = root_span_finishedAt

torch_client = TorchClient(**test_const.torch_credentials)


@pytest.fixture(scope="module", autouse=True)
def setup_torch_client():
    global torch_client
    torch_client = TorchClient(**test_const.torch_credentials)


@pytest.mark.order(1)
def test_create_pipeline_and_run():
    global pipeline_run_id, parent_span_context
    meta = PipelineMetadata(owner='sdk/pipeline-user', team='ADOC', codeLocation='...')
    pipeline_name_ = pipeline_name
    pp.pprint("explicit_pipeline_createdAt:" + str(explicit_pipeline_createdAt))
    print("explicit_pipeline_updatedAt:" + str(explicit_pipeline_updatedAt))
    pipeline = CreatePipeline(
        uid=pipeline_uid,
        name=pipeline_name_,
        description=f'The pipeline {pipeline_name_} has been created from acceldata-sdk using External integration',
        meta=meta,
        context={'pipeline_uid': pipeline_uid, 'pipeline_name': pipeline_name_},
        createdAt=explicit_pipeline_createdAt,
        updatedAt=explicit_pipeline_updatedAt
    )
    pipeline_res = torch_client.create_pipeline(pipeline=pipeline)
    pp.pprint(('Created pipeline with pipeline id :: ', pipeline_res.id))
    pp.pprint("explicit_pipeline_run_startedAt:" + str(explicit_pipeline_run_startedAt))
    pipeline_run = pipeline_res.create_pipeline_run(startedAt=explicit_pipeline_run_startedAt)
    global pipeline_run_id
    pipeline_run_id = pipeline_run.id
    span_name_ = f'{pipeline_uid}.root.span'
    global parent_span_context
    parent_span_context = pipeline_run.create_span(uid=span_name_, with_explicit_time=True)
    pp.pprint("===Starting the root span for the pipeline run=====")
    parent_span_context.start(created_at=root_span_created_at)


# Create job and span in torch and return the handle of span (not bounded by span)
def create_job_span_not_bounded(job_uid, inputs, outputs, metadata, context_job, span_uid, with_explicit_time):
    pp.pprint(
        "=================entering create_job_span_not_bounded for the job with uid " + job_uid + "==================")
    span_uid_temp = span_uid
    pipeline = torch_client.get_pipeline(pipeline_uid)
    pipeline_run = pipeline.get_run(pipeline_run_id)
    try:
        job = CreateJob(
            uid=job_uid,
            name=f'{job_uid} Job',
            pipeline_run_id=pipeline_run.id,
            description=f'{job_uid} created using torch job decorator',
            inputs=inputs,
            outputs=outputs,
            meta=metadata,
            context=context_job,
            with_explicit_time=with_explicit_time
        )
        job = pipeline.create_job(job)
        print("Create job response: " + str(job))
    except Exception as e:
        pp.pprint("Error in creating job")
        exception = e.__dict__
        pp.pprint(exception)
        raise e
    else:
        pp.pprint("Successfully created job." + job_uid)
        parent_span_context1 = pipeline_run.get_root_span()

        associated_job_uids = [job_uid]
        if span_uid is None:
            span_uid_temp = job_uid
        span_context = parent_span_context1.create_child_span(
            uid=span_uid_temp,
            context_data={
                'time': str(datetime.now())
            },
            associatedJobUids=associated_job_uids,
            with_explicit_time=True)
    return span_context


@pytest.mark.order(2)
def test_write_file_func():
    context_job = {'job': 'data_gene', 'uid': 'customers.data-generation', 'operator': 'write_file_func'}
    span_context_parent = create_job_span_not_bounded(
        job_uid='customers.data-generation',
        inputs=[],
        outputs=[Node(job_uid='customers.s3-upload')],
        metadata=JobMetadata('Jason', 'COKE', 'https://github.com/coke/reports/customers.kt'),
        context_job=context_job,
        span_uid='customers.data.generation',
        with_explicit_time=True)
    pp.pprint("entering write_file_func")
    size = csv_size
    span_context_parent.start(created_at=child_span_1_created_at)
    span_context_parent.send_event(GenericEvent(
        context_data={'Size': size - 1, 'total_file': 1,
                      'schema': 'name,address,dept_id'},
        event_uid="customers.data.generation.metadata",
        created_at=child_span_1_created_at))
    span_context_parent.send_event(LogEvent(
        context_data={'Size': size - 1, 'total_file': 1,
                      'schema': 'name,address,dept_id'},
        log_data="Customer data generated successfully.",
        created_at=child_span_1_created_at))
    # Ends span at the end of task
    span_context_parent.end(created_at=child_span_1_finished_at)
    # Upload generated file to s3 bucket


@pytest.mark.order(3)
def test_upload_file_to_s3():
    context_job = {'job': 'data_upload', 'time': str(datetime.now()), 'uid': 'customers.s3-upload',
                   'operator': 'upload_file_to_s3'}
    span_context_parent = create_job_span_not_bounded(
        job_uid='customers.s3-upload',
        inputs=[Node(job_uid='customers.data-generation')],
        outputs=[Node(asset_uid=f'{s3_ds}.{s3_customer}')],
        context_job=context_job,
        metadata=JobMetadata('Jason', 'COKE', 'https://github.com/coke/reports/customers.kt'),
        span_uid='customers.s3.upload',
        with_explicit_time=True)
    pp.pprint("entering upload_file_to_s3")
    span_context_parent.start(created_at=child_span_2_created_at)
    # End span at the end of task
    span_context_parent.end(created_at=child_span_2_finished_at)


@pytest.mark.order(5)
def test_rds_and_s3_clubbing():
    context_job = {'job': 'data_clubbing', 'time': str(datetime.now()), 'uid': 'customers.s3-postgres-clubbing',
                   'operator': 'rds_and_s3_clubbing'}

    # Child 3
    pp.pprint(" ======== Creating Child 3 on the root span =====================")
    parent_span_context1 = create_job_span_not_bounded(
        job_uid='customers.s3-postgres-clubbing',
        inputs=[Node(asset_uid=f'{s3_ds}.{s3_customer}'), Node(asset_uid=f'{postgres_ds}.{pg_table}')],
        outputs=[Node(asset_uid=f'{s3_ds}.{s3_postgres_customers}')],
        context_job=context_job,
        metadata=JobMetadata('BEN', 'COKE', 'https://github.com/coke/reports/rds_customers.kt'),
        span_uid='customers.s3.postgres.clubbing',
        with_explicit_time=True)
    parent_span_context1.start(created_at=child_span_3_created_at)
    pp.pprint("entering csv_pg_s3")
    pp.pprint("reading csv file")

    # Child 3->1
    pp.pprint(" ======== Creating Child 3 -> 1 =====================")
    read_csv_span = parent_span_context1.create_child_span(
        uid="read_csv",
        with_explicit_time=True
    )

    # 3->1 start
    pp.pprint(" ======== Starting Child 3 -> 1 =====================")
    read_csv_span.start(created_at=child_span_3_child_1_created_at)
    read_csv_span.send_event(
        GenericEvent(
            context_data={
                'total_file_to_be_read': 1, 'RDS_TABLES': '1',
            },
            event_uid="read_csv_data",
            created_at=child_span_3_child_1_created_at
        )
    )
    # 3->1 end
    pp.pprint(" ======== Ending Child 3 -> 1 =====================")
    read_csv_span.end(
        created_at=child_span_3_child_1_finished_at
    )

    # 3-> Generic event
    pp.pprint(" ======== Generic event Child 3 =====================")

    parent_span_context1.send_event(GenericEvent(
        context_data={'total_file_to_be_read': 1, 'RDS_TABLES': '1',
                      'DATA_INSERTED': 100, 'RDS_USER': 'postgres'},
        event_uid="customers.rds.migration.metadata",
        created_at=child_span_3_child_1_finished_at))

    # 3->2
    pp.pprint(" ======== Creating Child 3 -> 2 =====================")
    rds_merge_span = parent_span_context1.create_child_span(
        uid="rds_merge_data",
        with_explicit_time=True
    )

    # 3->2 start
    pp.pprint(" ======== Starting Child 3 -> 2 =====================")
    rds_merge_span.start(
        created_at=child_span_3_child_2_created_at
    )

    # 3->2->1
    pp.pprint(" ======== Creating Child 3 -> 2 -> 1 =====================")
    rds_dataframe_span = rds_merge_span.create_child_span(
        uid="rds_prepare_dataframe",
        with_explicit_time=True
    )

    # 3->2->1 start
    pp.pprint(" ======== Starting Child 3 -> 2 -> 1 =====================")
    rds_dataframe_span.start(
        created_at=child_span_3_child_2_child_1_created_at
    )

    # 3->2->1 end
    pp.pprint(" ======== Ending Child 3 -> 2 -> 1 =====================")
    rds_dataframe_span.end(
        created_at=child_span_3_child_2_child_1_finished_at
    )

    # 3->2->2
    pp.pprint(" ======== Creating Child 3 -> 2 -> 2 =====================")
    rds_save_dataframe_span = rds_merge_span.create_child_span(
        uid="rds_save_dataframe",
        context_data={'client_time': str(datetime.now())},
        with_explicit_time=True
    )

    # 3->2->2 start
    pp.pprint(" ======== Starting Child 3 -> 2 -> 2 =====================")
    rds_save_dataframe_span.start(
        context_data={'client_time': str(datetime.now())},
        created_at=child_span_3_child_2_child_2_created_at
    )

    # 3->2->2 end
    pp.pprint(" ======== Ending Child 3 -> 2 -> 2 =====================")
    rds_save_dataframe_span.end(
        context_data={'client_time': str(datetime.now())},
        created_at=child_span_3_child_2_child_2_finished_at
    )

    # 3->2 end
    pp.pprint(" ======== Ending Child 3 -> 2 =====================")
    rds_merge_span.end(
        created_at=child_span_3_child_2_finished_at
    )

    # 3->3
    pp.pprint(" ======== Creating Child 3 -> 3 =====================")
    write_csv_span = parent_span_context1.create_child_span(
        uid="write_csv",
        with_explicit_time=True
    )

    # 3->3 start
    pp.pprint(" ======== Starting Child 3 -> 3 =====================")
    write_csv_span.start(
        created_at=child_span_3_child_3_created_at
    )
    pp.pprint("Sending generic event for child 3->3")
    write_csv_span.send_event(
        GenericEvent(
            context_data={
                'client_time': str(datetime.now()),
                'total_file_to_be_written': 1, 'RDS_TABLES': '1'
            },
            event_uid="write_csv_data",
            created_at=child_span_3_child_3_created_at
        )
    )
    # 3->3 end
    pp.pprint(" ======== Ending Child 3 -> 3 =====================")
    write_csv_span.end(
        context_data={'client_time': "dummy"},
        created_at=child_span_3_child_3_finished_at
    )

    # End the torch span at the end of task
    pp.pprint(" ======== Ending Child 3 =====================")
    parent_span_context1.end(created_at=child_span_3_finished_at)


@pytest.mark.order(5)
def test_close_pipeline_run():
    # teardown: End the pipeline
    pipeline = torch_client.get_pipeline(test_const.pipeline_uid)
    pipeline_run = pipeline.get_run(pipeline_run_id)
    parent_span = pipeline_run.get_root_span()
    parent_span.end(context_data={'dag_status': 'SUCCESS', 'time': str(datetime.now())})
    pipeline_run.update_pipeline_run(
        context_data={'status': 'success'},
        result=PipelineRunResult.SUCCESS,
        status=PipelineRunStatus.COMPLETED
    )



from dataclasses import asdict

from acceldata_sdk.errors import TorchSdkException
from acceldata_sdk.models.profile import ProfilingType, JobType, Profile, StartProfilingRequest
from acceldata_sdk.models.assetType import AssetType
from acceldata_sdk.models.create_asset import RelationType, CreateAssetRelation, AssetMetadata
from acceldata_sdk.models.datasource import DataSource, SourceType
from acceldata_sdk.models.tags import AssetLabel, CustomAssetMetadata
from acceldata_sdk.models.common_types import ExecutionType


from enum import Enum, auto
from typing import List
import logging

logger = logging.getLogger('asset')
logger.setLevel(logging.INFO)

class Metadata:

    def __init__(self, assetId, createdAt=None, updatedAt=None, currentSnapshot=None, id=None, items=None, metaDataHash=None, snapshots=None, *args, **kwargs):
        self.assetId = assetId
        self.createdAt = createdAt
        self.updatedAt = updatedAt
        self.currentSnapshot = currentSnapshot
        self.id = id
        self.metaDataHash = metaDataHash
        if isinstance(items, list):
            items_list = list(items)
            asset_metadata = []
            for i in items_list:
                asset_mt = AssetMetadata(**i)
                asset_metadata.append(asset_mt)
            self.items = asset_metadata
        else:
            self.items = items
        if isinstance(snapshots, list):
            self.snapshots = list(snapshots)
        else:
            self.snapshots = snapshots

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"Metadata({self.__dict__})"


class CustomAssetType(Enum):
    SQL_VIEW = auto()
    SQL_VIEW_COLUMN = auto()
    VISUAL_VIEW = auto()
    VISUAL_VIEW_COLUMN = auto()


class Asset:
    def __init__(self,
                 alias=None,
                 assembly=None,
                 assetType: AssetType = None,
                 createdAt=None,
                 currentSnapshot=None,
                 description=None,
                 id=None,
                 isCustom=None,
                 isDeleted=None,
                 name=None,
                 parentId=None,
                 snapshots=None,
                 sourceType: SourceType = None,
                 uid=None,
                 updatedAt=None,
                 client=None,
                 customAssetType: CustomAssetType = None,
                 assemblyId=None,
                 isSegmented=None,
                 *args, **kwargs
                 ):
        """
            Description:
                Asset class
        :param alias: alias of the asset
        :param assembly: (Datasource) data source details
        :param assetType: type of the asset
        :param createdAt: creation time of the asset
        :param currentSnapshot: current version of the asset
        :param description: desc of the asset
        :param id: asset id
        :param isCustom: is custom asset or not
        :param isDeleted: is deleted or not in current version of the datasource
        :param name: name of the asset
        :param parentId: parent id of the asset
        :param snapshots: version list
        :param sourceType: source type of the asset's datasource
        :param uid: uid of the asset
        :param updatedAt: updated time of the asset
        """
        self.alias = alias
        self.createdAt = createdAt
        self.currentSnapshot = currentSnapshot
        self.description = description
        self.id = id
        self.isCustom = isCustom
        self.isDeleted = isDeleted
        self.assemblyId = assemblyId
        self.isSegmented = isSegmented
        self.name = name
        self.parentId = parentId
        self.snapshots = snapshots
        self.uid = uid
        self.updatedAt = updatedAt
        if isinstance(assembly, dict):
            self.datasource = DataSource(**assembly)
        else:
            self.datasource = assembly
        if isinstance(assetType, dict):
            self.assetType = AssetType(**assetType)
        else:
            self.assetType = assetType
        if isinstance(sourceType, dict):
            self.sourceType = SourceType(**sourceType)
        else:
            self.sourceType = sourceType
        if isinstance(customAssetType, dict):
            self.customAssetType = CustomAssetType(**customAssetType)
        else:
            self.customAssetType = customAssetType
        self.client = client

    def __repr__(self):
        return f"Asset({self.__dict__})"

    def get_labels(self):
        return self.client.get_asset_labels(self.id)

    def add_labels(self, labels: List[AssetLabel]):
        lbls = []
        for label in labels:
            label_dict = {
                'key': label.key,
                'value': label.value
            }
            lbls.append(label_dict)
        payload = {
            'labels': lbls
        }
        return self.client.add_asset_labels(self.id, payload)

    def get_tags(self):
        return self.client.get_asset_tags(self.id)

    def add_tag(self, tag):
        payload = {
            "name": tag
        }
        return self.client.add_asset_tag(self.id, payload)

    def get_metadata(self):
        return self.client.get_asset_metadata(self.id)

    def add_custom_metadata(self, custom_metadata: List[CustomAssetMetadata]):
        """
        Description:
            Add custom metadata
        :param custom_metadata: list of CustomAssetMetadata Object
        :return:
        """
        metadts = dict()
        for custom_metadatum in custom_metadata:
            metadts[custom_metadatum.key] = custom_metadatum.value
        return self.client.add_custom_metadata(self.id, metadts)

    def start_profile(self, profiling_type: ProfilingType = None,
                      start_profiling_request: StartProfilingRequest = None) -> Profile:
        payload = {}
        logger.info(
            f"Invoked start_profile method of the Asset\n"
            f"ProfilingType: {profiling_type}\n"
            f"StartProfilingRequest: {start_profiling_request.to_dict() if start_profiling_request else 'None'}"
        )

        if start_profiling_request is None:
            logger.info(
                "StartProfilingRequest is None , executing profile based on profiling type")
            payload = {
                'data': {
                    'profilingType': profiling_type.value
                }
            }

        if start_profiling_request is not None:
            logger.info(f"Marker config: {start_profiling_request.markerConfig}")
            if start_profiling_request.profilingType == ExecutionType.SELECTIVE and not start_profiling_request.markerConfig:
                raise TorchSdkException(
                    f'markerConfig is a mandatory parameter for execution type {ExecutionType.SELECTIVE}')
            else:
                logger.info("Converting start_profiling_request to dictionary")
                payload = start_profiling_request.to_dict()
                logger.info(f"StartProfilingRequest to dictionary: {payload}")
        logger.info(f"Executing Profiling with the following payload: {payload}")
        return self.client.profile_asset(self.id, payload)

    def get_latest_profile_status(self):
        return self.client.get_latest_profile_status(self.id)

    def sample_data(self):
        return self.client.sample_data(self.id)



import time
import pytest
import logging

LOGGER = logging.getLogger(__name__)


def retry_operation(operation, max_retries, retry_interval, *args, **kwargs):
    retry_count = 0
    LOGGER.info("Retry count: %s", retry_count)
    while retry_count < max_retries:
        try:
            LOGGER.info("Executing operation...")
            result = operation(*args, **kwargs)
            LOGGER.info(f"Operation result: {result}")
            assert result is not None
            return result
        except Exception as e:
            LOGGER.error(str(e))
            retry_count += 1
            if retry_count < max_retries:
                retry_interval_minutes = retry_interval // 60
                LOGGER.info(f"Retrying in {retry_interval_minutes} minutes...")
                time.sleep(retry_interval)
            else:
                pytest.fail("Operation failed after multiple retries.")
    return None



import pytest
import pprint
from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_sdk.events.log_events import LogEvent
from acceldata_sdk.models.job import JobMetadata, Node, CreateJob
from acceldata_sdk.models.pipeline import CreatePipeline, PipelineMetadata, PipelineRunResult, PipelineRunStatus
from acceldata_sdk.torch_client import TorchClient
from datetime import datetime
import test_constants as test_const

pp = pprint.PrettyPrinter(indent=4)

# Global variables needed across functions
parent_span_context = None
pipeline_run_id = None

torch_client = TorchClient(**test_const.torch_credentials)

@pytest.fixture(scope="module", autouse=True)
def setup_pipeline_and_run():
    global parent_span_context, pipeline_run_id, torch_client
    pp.pprint("================= entering create_pipeline_and_run ==================")
    meta = PipelineMetadata(owner='sdk/pipeline-user', team='ADOC', codeLocation='...')
    pipeline_name_ = test_const.pipeline_name
    pipeline = CreatePipeline(
        uid=test_const.pipeline_uid,
        name=pipeline_name_,
        description=f'The pipeline {pipeline_name_} has been created from acceldata-sdk using External integration',
        meta=meta,
        context={'pipeline_uid': test_const.pipeline_uid, 'pipeline_name': pipeline_name_}
    )
    pipeline_res = torch_client.create_pipeline(pipeline=pipeline)
    print('Created pipeline with pipeline id :: ', pipeline_res.id)

    pipeline_run = pipeline_res.create_pipeline_run()
    pipeline_run_id = pipeline_run.id
    span_name_ = f'{test_const.pipeline_uid}.root.span'
    parent_span_context = pipeline_run.create_span(uid=span_name_)
    print("===Starting the root span for the pipeline run=====")
    parent_span_context.start()
    yield
    # teardown: End the pipeline
    pipeline = torch_client.get_pipeline(test_const.pipeline_uid)
    pipeline_run = pipeline.get_run(pipeline_run_id)
    parent_span = pipeline_run.get_root_span()
    parent_span.end(context_data={'dag_status': 'SUCCESS', 'time': str(datetime.now())})
    pipeline_run.update_pipeline_run(
        context_data={'status': 'success'},
        result=PipelineRunResult.SUCCESS,
        status=PipelineRunStatus.COMPLETED
    )

def create_job_span_not_bounded(job_uid, inputs, outputs, metadata, context_job, span_uid):
    print(
        "=================entering create_job_span_not_bounded for the job with uid " + job_uid + "==================")
    span_uid_temp = span_uid
    pipeline = torch_client.get_pipeline(test_const.pipeline_uid)
    pipeline_run = pipeline.get_run(pipeline_run_id)
    try:
        job = CreateJob(
            uid=job_uid,
            name=f'{job_uid} Job',
            pipeline_run_id=pipeline_run.id,
            description=f'{job_uid} created using torch job decorator',
            inputs=inputs,
            outputs=outputs,
            meta=metadata,
            context=context_job
        )
        job = pipeline.create_job(job)
        pp.pprint("Create job response: " + str(job))
    except Exception as e:
        print("Error in creating job")
        exception = e.__dict__
        print(exception)
        raise e
    else:
        pp.pprint("Successfully created job." + job_uid)
        parent_span_context1 = pipeline_run.get_root_span()

        associated_job_uids = [job_uid]
        if span_uid is None:
            span_uid_temp = job_uid
        span_context = parent_span_context1.create_child_span(
            uid=span_uid_temp,
            context_data={
                'time': str(datetime.now())
            },
            associatedJobUids=associated_job_uids)
    return span_context

@pytest.mark.order(1)
def test_write_file_func():
    pp.pprint("================= entering write_file_func ==================")
    context_job = {'job': 'data_gene', 'uid': 'customers.data-generation', 'operator': 'write_file_func'}
    span_context_parent = create_job_span_not_bounded(
        job_uid='customers.data-generation',
        inputs=[],
        outputs=[Node(job_uid='customers.s3-upload')],
        metadata=JobMetadata('Jason', 'COKE', 'https://github.com/coke/reports/customers.kt'),
        context_job=context_job,
        span_uid='customers.data.generation')
    pp.pprint("entering write_file_func")
    size = test_const.csv_size
    span_context_parent.start()
    span_context_parent.send_event(GenericEvent(
        context_data={'Size': size - 1, 'total_file': 1, 'schema': 'name,address,dept_id'},
        event_uid="customers.data.generation.metadata"))
    span_context_parent.send_event(LogEvent(
        context_data={'Size': size - 1, 'total_file': 1, 'schema': 'name,address,dept_id'},
        log_data="Customer data generated successfully."))
    span_context_parent.end()

@pytest.mark.order(2)
def test_upload_file_to_s3():
    context_job = {'job': 'data_upload', 'time': str(datetime.now()), 'uid': 'customers.s3-upload', 'operator': 'upload_file_to_s3'}
    span_context_parent = create_job_span_not_bounded(
        job_uid='customers.s3-upload',
        inputs=[Node(job_uid='customers.data-generation')],
        outputs=[Node(asset_uid=f'{test_const.s3_ds}.{test_const.s3_customer}')],
        context_job=context_job,
        metadata=JobMetadata('Jason', 'COKE', 'https://github.com/coke/reports/customers.kt'),
        span_uid='customers.s3.upload')
    pp.pprint("entering upload_file_to_s3")
    span_context_parent.start()
    span_context_parent.end()

@pytest.mark.order(3)
def test_rds_and_s3_clubbing():
    context_job = {'job': 'data_clubbing', 'time': str(datetime.now()), 'uid': 'customers.s3-postgres-clubbing', 'operator': 'rds_and_s3_clubbing'}

    pp.pprint(" ======== Creating Child 3 on the root span =====================")
    parent_span_context1 = create_job_span_not_bounded(
        job_uid='customers.s3-postgres-clubbing',
        inputs=[Node(asset_uid=f'{test_const.s3_ds}.{test_const.s3_customer}'), Node(asset_uid=f'{test_const.postgres_ds}.{test_const.pg_table}')],
        outputs=[Node(asset_uid=f'{test_const.s3_ds}.{test_const.s3_postgres_customers}')],
        context_job=context_job,
        metadata=JobMetadata('BEN', 'COKE', 'https://github.com/coke/reports/rds_customers.kt'),
        span_uid='customers.s3.postgres.clubbing')
    parent_span_context1.start()
    pp.pprint("entering csv_pg_s3")
    pp.pprint("reading csv file")

    pp.pprint(" ======== Creating Child 3 -> 1 =====================")
    read_csv_span = parent_span_context1.create_child_span(uid="read_csv")

    pp.pprint(" ======== Starting Child 3 -> 1 =====================")
    read_csv_span.start()
    read_csv_span.send_event(GenericEvent(context_data={'total_file_to_be_read': 1, 'RDS_TABLES': '1'}, event_uid="read_csv_data"))
    read_csv_span.end()

    pp.pprint(" ======== Generic event Child 3 =====================")
    parent_span_context1.send_event(GenericEvent(
        context_data={'total_file_to_be_read': 1, 'RDS_TABLES': '1', 'DATA_INSERTED': 100, 'RDS_USER': 'postgres'},
        event_uid="customers.rds.migration.metadata"))

    pp.pprint(" ======== Creating Child 3 -> 2 =====================")
    rds_merge_span = parent_span_context1.create_child_span(uid="rds_merge_data")

    pp.pprint(" ======== Starting Child 3 -> 2 =====================")
    rds_merge_span.start()

    pp.pprint(" ======== Creating Child 3 -> 2 -> 1 =====================")
    rds_dataframe_span = rds_merge_span.create_child_span(uid="rds_prepare_dataframe")

    pp.pprint(" ======== Starting Child 3 -> 2 -> 1 =====================")
    rds_dataframe_span.start()
    rds_dataframe_span.end()

    pp.pprint(" ======== Creating Child 3 -> 2 -> 2 =====================")
    rds_save_dataframe_span = rds_merge_span.create_child_span(uid="rds_save_dataframe", context_data={'client_time': str(datetime.now())})

    pp.pprint(" ======== Starting Child 3 -> 2 -> 2 =====================")
    rds_save_dataframe_span.start(context_data={'client_time': str(datetime.now())})
    rds_save_dataframe_span.end(context_data={'client_time': str(datetime.now())})

    pp.pprint(" ======== Ending Child 3 -> 2 =====================")
    rds_merge_span.end()

    pp.pprint(" ======== Creating Child 3 -> 3 =====================")
    write_csv_span = parent_span_context1.create_child_span(uid="write_csv")

    pp.pprint(" ======== Starting Child 3 -> 3 =====================")
    write_csv_span.start()
    print("Sending generic event for child 3->3")
    write_csv_span.send_event(GenericEvent(
        context_data={'client_time': str(datetime.now()), 'total_file_to_be_written': 1, 'RDS_TABLES': '1'},
        event_uid="write_csv_data"))
    write_csv_span.end(context_data={'client_time': "dummy"})

    pp.pprint(" ======== Ending Child 3 =====================")
    parent_span_context1.end()



import logging
import time
from abc import ABC
from typing import List
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

import requests
from acceldata_sdk.models.connection import ConnectionType, AnalyticsPipeline, Connection, ConnectionCheck, \
    ConnectionCheckStatus
from acceldata_sdk.models.profile import Profile, ProfileRequest
from acceldata_sdk.models.rule import Rule, RuleCancelResponse, RuleExecution, ExecutionResult
from acceldata_sdk.models.common_types import PolicyExecutionRequest, ExecutionType
from acceldata_sdk.models.tags import TAG, AssetTag, AssetLabel
from acceldata_sdk.errors import APIError, TorchSdkException
from acceldata_sdk.models.assetType import AssetType
from acceldata_sdk.models.datasource import DataSource, SourceType, CrawlerStatus
from acceldata_sdk.models.job import Job
from acceldata_sdk.models.pipeline import CreatePipeline, Pipeline, PipelineRun, PipelineInfo, PipelineDetails, PipelineSourceType
from acceldata_sdk.models.span import Span, SpanContextEvent, CreateSpanEvent
from acceldata_sdk.client import TorchClientInterFace
from acceldata_sdk.models.asset import Asset, Metadata
from requests_toolbelt import MultipartEncoder
from acceldata_sdk.models.dqrule import DataQualityExecutionResult, DataQualityRule, DataQualityRuleResource
from acceldata_sdk.models.reconcillationrule import ReconciliationExecutionResult, ReconciliationRuleResource
from acceldata_sdk.models.ruleExecutionResult import RuleResult, PolicyFilter, ExecutionPeriod, RuleType, \
    RuleExecutionSummary
from acceldata_sdk.constants import AssetSourceType, SdkSupportedVersions, TorchBuildVersion
from acceldata_sdk.common import Executor
from acceldata_sdk.constants import PolicyType, FailureStrategy, RuleExecutionStatus, \
    MIN_TORCH_BACKEND_VERSION_FOR_RULE_ID_API
from acceldata_sdk.models.rule import RuleResource
from semantic_version import Version, SimpleSpec
from acceldata_sdk.api_version_utils import APIVersionUtils
from acceldata_sdk.time_range_utils import TimeRangeCalculator

from typing import List
from dataclasses import asdict
import json
# import re
from datetime import datetime, timedelta

_HEADERS = {'User-Agent': 'Acceldata-sdk', 'accessKey': None, 'secretKey': None, 'Content-Type': 'application/json'}
catalog_api_path = "/catalog-server/api"
pipeline_api_path = "/torch-pipeline/api"
admin_api_path = "/admin/api"

import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class TorchHttpClient(TorchClientInterFace, ABC):
    access_key = None
    secret_key = None
    logger = logging.getLogger('torch')
    logger.setLevel(logging.INFO)

    # Create torch client by passing secret and access keys of catalog server for a given url
    def __init__(self, url, timeout_ms=10000, access_key: str = None, secret_key: str = None):
        """
            Description : Torch client is used to send data to catalog server.

            :param url: (String) url of the catalog server
                        ex:  https://torch.acceldata.local:5443
            :param timeout_ms: (Integer) timeout of the requests sending to catalog
            :param access_key: (String) Access key of API key. You can generate API key from torch ui's setting
            :param secret_key: (String) Secret key of API key.

            : TorchClient = TorchClient(url='https://torch.acceldata.local:5443', access_key='OY2VVIN2N6LJ', secret_key='da6bDBimQfXSMsyyhlPVJJfk7Zc2gs')
        """
        self.timeout_ms = timeout_ms
        self._timeout = 10000
        self._catalog_api_base = f"{url}{catalog_api_path}"
        self._pipeline_api_base = f"{url}{pipeline_api_path}"
        is_onprem = url.endswith('/torch')
        loc = url.rfind('/torch')
        if is_onprem:
            self._admin_api_base = f"{url[:loc]}{admin_api_path}"
        else:
            self._admin_api_base = f"{url}{admin_api_path}"

        if access_key is None and secret_key is None:
            raise Exception('Access key and secret key - required')
        self.access_key = access_key
        self.secret_key = secret_key
        _HEADERS['accessKey'] = access_key
        _HEADERS['secretKey'] = secret_key

    # # convert pipeline object to dict type
    def _convert_pipeline_to_dict(self, pipeline: CreatePipeline):
        """
        Description:
            Convert pipeline object to dict type.
        :param pipeline: createPipeline class instance
        :return: pipeline in dict type
        """
        pipeline_dict = {k: v for k, v in pipeline.__dict__.items() if v is not None}
        if pipeline.meta is not None:
            meta = asdict(pipeline.meta)
            pipeline_dict['meta'] = meta
            # Check if sourceType is an instance of PipelineSourceType, and convert it using to_dict
        if isinstance(pipeline.sourceType, PipelineSourceType):
            pipeline_dict['sourceType'] = pipeline.sourceType.to_dict()
        return pipeline_dict

    def get_supported_sdk_versions(self) -> SdkSupportedVersions:
        response = self._get(
            f'{self._catalog_api_base}/sdk/compatible-versions'
        )
        return SdkSupportedVersions(**response['versions'])

    def get_torch_version(self) -> TorchBuildVersion:
        response = self._get(
            f'{self._admin_api_base}/build-details'
        )
        return TorchBuildVersion(**response)

    # function to create pipeline
    def create_pipeline(self, pipeline: CreatePipeline) -> Pipeline:
        """
        Description:
            Used to create pipeline
        :param pipeline: createPipeline class instance that you want to create
        :return: pipeline class instance
        """
        payload = self._convert_pipeline_to_dict(pipeline)
        pipeline_payload = {'pipeline': payload}
        response = self._put(
            f'{self._pipeline_api_base}/pipelines',
            payload=pipeline_payload
        )
        response['pipeline']['client'] = self
        self.logger.info('Pipeline Created')
        return Pipeline(**response['pipeline'])

    # function to delete pipeline
    def delete_pipeline(self, id):
        """
        Description:
            Used to delete pipeline
        :param id: id of pipeline
        """

        response = self._delete(
            f'{self._pipeline_api_base}/pipelines/{id}'
        )
        self.logger.info(f'Pipeline {id} deleted')
        return response

    # get pipeline by uid or id
    def get_pipeline(self, pipeline_identity) -> Pipeline:
        """
            Description:
                To get an existing pipeline from torch catalog
        :param pipeline_identity: uid or id of the pipeline
        :return:(Pipeline) pipeline class instance
        """
        url = f'{self._pipeline_api_base}/pipelines/{pipeline_identity}'
        response = self._get(url)
        response['pipeline']['client'] = self
        return Pipeline(**response['pipeline'])

    # get all pipelines
    def get_pipelines(self) -> List[PipelineInfo]:
        """
            Description:
                To get all pipelines from torch catalog
        :return:(List[PipelineInfo]) PipelineInfo class instance
        """
        url = f'{self._pipeline_api_base}/pipelines'
        response = self._get(url)
        pipelines_infos = list()
        pipelines = response['pipelines']
        for obj in pipelines:
            obj['pipeline']['client'] = self
            pipelines_infos.append(PipelineInfo(pipeline=obj['pipeline'], runs=obj['runs']))

        return pipelines_infos

    # to create job for any given pipeline
    def create_job(self, job: {}, pipelineId: int) -> Job:
        """
        Description:
            Used to create job in a pipeline
        :param pipelineId: pipeline id of the respective pipeline
        :param job: createJob class instance that you want to add in pipeline
        :return: Job class instance of created job
        """
        response = self._put(
            f'{self._pipeline_api_base}/pipelines/{pipelineId}/jobs',
            payload=job
        )
        self.logger.info('Pipeline Job Created')
        return Job(**response['node'])

    def delete_datasource(self, datasource_id: int):
        """
        Description:
            used to delete datasource
        :param datasource_id: id of the datasource to be updated
        :return: response of the operation
        """
        response = self._delete(
            f'{self._catalog_api_base}/assemblies/{datasource_id}'
        )
        return response

    def start_crawler(self, datasource_name) -> CrawlerStatus:
        response = self._post(
            f'{self._catalog_api_base}/crawler/{datasource_name}',
            payload={}
        )
        return CrawlerStatus(**response['data'])

    def get_crawler_status(self, datasource_name) -> CrawlerStatus:
        response = self._get(
            f'{self._catalog_api_base}/crawler/{datasource_name}',
        )
        return CrawlerStatus(**response['data'])

    # create run for a pipeline
    def create_pipeline_run(self, pipeline_run: {}) -> PipelineRun:
        """
        Description:
            used to create a pipeline run
        :param pipeline_run:
        :return: pipelineRun class instance
        """
        pipelineId = pipeline_run['run']['pipelineId']
        response = self._post(
            f'{self._pipeline_api_base}/pipelines/{pipelineId}/runs',
            payload=pipeline_run
        )
        response['run']['client'] = self
        self.logger.info('Pipeline Run Created')
        return PipelineRun(**response['run'])

    # update run for a pipeline
    def update_pipeline_run(self, pipeline_run_id: int, pipeline_run: {}) -> PipelineRun:
        """
        Description:
            used to update an existing pipeline run
        :param pipeline_run_id: pipeline run id that you want to update
        :param pipeline_run: pipelineRun class instance that you want to update
        :return: updated pipelineRun class instance
        """
        response = self._put(
            f'{self._pipeline_api_base}/pipelines/runs/{pipeline_run_id}',
            payload=pipeline_run
        )
        response['run']['client'] = self
        self.logger.info('Pipeline Run Updated')
        return PipelineRun(**response['run'])

    # get latest pipeline run
    def get_latest_pipeline_run(self, pipeline_id: int) -> PipelineRun:
        """
            Description:
                To get latest pipeline run instance of any pipeline
        :param pipeline_id: id of the pipeline
        :return: PipelineRun instance
        """
        url = f'{self._pipeline_api_base}/pipelines/{pipeline_id}/latestRun'
        response = self._get(url)
        response['run']['client'] = self
        return PipelineRun(**response['run'])

    # get pipeline run with particular continuation id or run id
    def get_pipeline_run(self, pipeline_run_id: int = None, continuation_id: str = None,
                         pipeline_id: str = None) -> PipelineRun:
        """
            Description:
                To get pipeline run with particular id or continuation id of any pipeline
        :param pipeline_run_id: run id of the pipeline run
        :param continuation_id: continuation id of the pipeline run
        :param pipeline_id: id of the pipeline. This is a mandatory parameter when run is being queried using continuation_id
        :return: PipelineRun instance
        """
        if pipeline_run_id is not None:
            url = f'{self._pipeline_api_base}/pipelines/runs/{pipeline_run_id}'
            response = self._get(url)
            return_run = response['run']
        elif continuation_id is not None and pipeline_id is not None:
            url = f'{self._pipeline_api_base}/pipelines/{pipeline_id}/runs'
            payload = {"continuationId": continuation_id}
            response = self._get(url, params=payload)
            if response["runs"] is not None and len(response["runs"]) > 0:
                return_run = response["runs"][0]
            else:
                raise TorchSdkException('No runs found against the given continuation_id.')
        else:
            raise TorchSdkException('Please provide either continuation_id and pipeline_id or pipeline_run_id')

        return_run['client'] = self
        return PipelineRun(**return_run)

    # get pipeline runs for a pipeline
    def get_pipeline_runs(self, pipeline_id: int) -> List[PipelineRun]:
        """
            Description:
                To get pipeline run  id of any pipeline
        :param pipeline_id: id of the pipeline
        :return: List[PipelineRun] instance
        """
        url = f'{self._pipeline_api_base}/pipelines/{pipeline_id}/runs'
        response = self._get(url)
        runs = list(response['runs'])
        runs_list = []
        for res in runs:
            res['client'] = self
            runs_list.append(PipelineRun(**res))
        return runs_list

    # create span for pipeline run
    def create_span(self, pipeline_run_id: int, span: dict) -> Span:
        """
        Description:
            used to create span for any pipeline run
        :param pipeline_run_id:
        :param span: Span class instance
        :return: Span
        """
        response = self._post(
            f'{self._pipeline_api_base}/pipelines/runs/{pipeline_run_id}/spans',
            payload=span
        )
        self.logger.info('Span Created')
        return Span(**response['span'])

    def get_child_spans(self, span_id: int):
        response = self._get(
            f'{self._pipeline_api_base}/pipelines/spans/{span_id}/childSpans'
        )
        self.logger.info('Fetched child spans.')
        child_spans = list(response['childSpans'])
        child_spans_res = []
        for res in child_spans:
            child_spans_res.append(Span(**res))
        return child_spans_res

    def get_span(self, pipeline_run_id: int, uid: str) -> Span:
        """
            Description:
                Get span of the pipeline run by uid
        :param pipeline_run_id: pipeline run id
        :param uid: uid of the span
        :return: SpanContext instance of the input span uid
        """
        url = f'{self._pipeline_api_base}/pipelines/runs/{pipeline_run_id}/spans/{uid}'
        response = self._get(url)
        response['span']['client'] = self
        return Span(**response['span'])

    def get_root_span(self, pipeline_run_id: int) -> Span:
        """
            Description:
                Get root span of the pipeline run
        :param pipeline_run_id: pipeline run id
        :return: SpanContext instance of the root span
        """
        url = f'{self._pipeline_api_base}/pipelines/runs/{pipeline_run_id}/spans'
        params = {"onlyRootSpan": "true"}
        response = self._get(url, params=params)
        return_value = dict()
        return_value['span'] = dict()
        return_value['span']['client'] = self
        if response['spans']:
            return_value['span'] = response['spans'][0]
            return_value['span']['client'] = self
            return Span(**return_value['span'])
        else:
            return None

    def get_spans(self, pipeline_run_id):
        response = self._get(
            f'{self._pipeline_api_base}/pipelines/runs/{pipeline_run_id}/spans'
        )
        self.logger.info('Fetched spans.')
        spans = list(response['spans'])
        spans_res = []
        for res in spans:
            spans_res.append(Span(**res))
        return spans_res

    # convert span object to dict type
    def convert_span_event_to_dict(self, span_event: CreateSpanEvent):
        """
            Description:
                Convert CreateSpanEvent class instance to dict type
            :param spanEvent: CreateSpanEvent class instance
            :return: dict form of CreateSpanEvent class instance
        """
        payload = span_event.__dict__
        # payload['eventUid'] = span_event.eventType
        event_payload = {'event': payload}
        return event_payload

    # create span event for any span
    def create_span_event(self, span_event: CreateSpanEvent) -> SpanContextEvent:
        """
        Description:
            used to create span event
        :param span_event: CreateSpanEvent class instance that you want to create
        :return: SpanContextEvent class instance
        """
        payload = self.convert_span_event_to_dict(span_event)
        if span_event.spanId is None:
            Exception('To update a pipeline run, id is required.')
        response = self._post(
            f'{self._pipeline_api_base}/pipelines/spans/{span_event.spanId}/events',
            payload=payload
        )
        self.logger.info('Span Event Created')
        return SpanContextEvent(**response['event'])

    def get_run_details(self, pipeline_id: int, version_id: int):
        """
            Description:
                Get details of the pipeline by versionid
        :param pipeline_id: pipeline  id
        :param version_id: version_id of the pipeline
        :return: SpanContext instance of the input span uid
        """
        url = f'{self._pipeline_api_base}/pipelines/{pipeline_id}/details?versionId={version_id}'
        response = self._get(url)
        details = response['details']
        details['pipeline']['client'] = self
        return PipelineDetails(**details)

    def get_all_asset_types(self):
        """
        Description:
            get all asset types supported in torch xatalog
        :return: list of asset types
        """
        response = self._get(
            f'{self._catalog_api_base}/asset-types'
        )
        asset_types = list(response['data'])
        asset_types_res = []
        for res in asset_types:
            asset_types_res.append(AssetType(**res))
        return asset_types_res

    def get_all_source_types(self):
        """
        Description:
            get all source types supported in torch catalog
        :return: list of all source type
        """
        response = self._get(
            f'{self._catalog_api_base}/source-types'
        )
        source_types = list(response['data'])
        source_types_res = []
        for res in source_types:
            source_types_res.append(SourceType(**res))
        return source_types_res

    def get_connection_types(self):
        """
        Description:
            get all source types supported in torch catalog
        :return: list of all source type
        """
        response = self._get(
            f'{self._catalog_api_base}/connection-types'
        )
        connection_types = list(response['data'])
        conn_type_response = []
        for res in connection_types:
            conn_type_response.append(ConnectionType(**res))
        return conn_type_response

    def get_tags(self):
        response = self._get(
            f'{self._catalog_api_base}/assets/tags'
        )
        tags = list(response['tags'])
        tags_list = []
        for res in tags:
            tags_list.append(TAG(**res))
        return tags_list

    def get_datasource(self, assembly_identifier, properties: bool = False) -> DataSource:
        """
        Description:
            Find datasource by its name or id in torch catalog
        :param assembly_identifier: name or id of the datasource given in torch
        :param properties: optional parameter, bool, to get datasource properties as well
        :return: (DataSource) datasource
        """
        url = f'{self._catalog_api_base}/assemblies/{assembly_identifier}?properties={properties}'

        response = self._get(url)
        datasource = response['data']
        if len(datasource) > 0:
            datasource['client'] = self
            return DataSource(**datasource)
        raise Exception('Datasource not found.')

    def get_datasources(self, type: AssetSourceType = None) -> List[DataSource]:
        """
        Description:
            Find datasources by its type in torch catalog
        :param type: type of the datasource given in torch, optional
        :return: list(DataSource) datasource
        """
        url = f'{self._catalog_api_base}/assemblies'
        if type is not None:
            params = dict()
            params['sourceType'] = type.name
            response = self._get(url, params=params)
        else:
            response = self._get(url)
        datasources = list(response['data'])
        ds_list = []
        for res in datasources:
            res['client'] = self
            ds_list.append(DataSource(**res))
        return ds_list
        raise Exception('Datasource not found.')

    def get_all_datasources(self):
        """
        Description:
            list all datasources in torch catalog
        :return: (DataSource) list of datasource
        """
        url = f'{self._catalog_api_base}/assemblies'
        response = self._get(url)
        datasources = list(response['data'])
        ds_list = []
        for res in datasources:
            ds_list.append(DataSource(**res))
        return ds_list

    def get_asset(self, identifier):
        """"
            Description:
                Find an asset of the datasource
            :param identifier: uid or ID of the asset
        """
        if check_int(identifier):
            return self.get_asset_by_id(id=identifier)
        else:
            return self.get_asset_by_uid(uid=identifier)

    def get_asset_by_id(self, id: int) -> Asset:
        """
        Description:
            used to find an asset
        :param id: id of an asset
        :return: asset data (Asset)
        """
        if id is None:
            raise Exception('Asset id is required')
        url = f'{self._catalog_api_base}/assets?ids={id}'
        asset = self._get_asset(url)
        asset_res = list(asset['data'])[0]
        asset_res['client'] = self
        return Asset(**asset_res)

    def get_asset_by_uid(self, uid: str):
        """
        Description:
            used to find an asset
        :param uid: uid of an asset
        :return: asset data (Asset)
        """
        if uid is None:
            raise Exception('Asset uid is required')
        url = f'{self._catalog_api_base}/assets?uid={uid}'
        asset = self._get_asset(url)
        if asset is None:
            return None
        asset_data = asset['data']
        if asset_data is None or not asset_data or len(asset_data) == 0:
            return None
        asset_res = list(asset_data)[0]
        asset_res['client'] = self
        return Asset(**asset_res)

    def _get_asset(self, url: str):
        response = self._get(
            url=url
        )
        return response

    def get_asset_metadata(self, id: int):
        url = f'{self._catalog_api_base}/assets/{id}/metadata'
        response = self._get(
            url=url
        )
        metadata = Metadata(**response['data'])
        return metadata

    def add_custom_metadata(self, id: int, custom_metadata: dict):
        url = f'{self._catalog_api_base}/assets/{id}/metadata/custom'
        response = self._post(
            url=url,
            payload=custom_metadata
        )
        metadata = Metadata(**response['data'])
        return metadata

    def profile_asset(self, id: int, payload: dict) -> Profile:
        url = f'{self._catalog_api_base}/assets/{id}/profile'
        response = self._post(
            url=url,
            payload=payload
        )
        response['data']['client'] = self
        profile_response = Profile(**response['data'])
        return profile_response

    def sample_data(self, asset_id: int):
        url = f'{self._catalog_api_base}/assets/{asset_id}/sample/async'
        response = self._post(url=url, as_json=True, payload={})

        # This is a file/kafka based asset sampling
        if response is not None and response.get('requestId') is not None:
            request_id = response['requestId']
            sample_data_result = self.get_sample_data_result(request_id)

            while sample_data_result['jobExecutionStatus'] == 'IN_PROGRESS':
                time.sleep(5)
                sample_data_result = self.get_sample_data_result(request_id)

            return sample_data_result
        else:
            return response

    def get_sample_data_result(self, request_id: str):
        url = f'{self._catalog_api_base}/assets/sample/result/{request_id}'
        response = self._get(
            url=url
        )
        return response

    def get_profile_request_details(self, asset_id: int, req_id: int):
        url = f'{self._catalog_api_base}/assets/{asset_id}/profile/{req_id}'
        response = self._get(
            url=url
        )
        return response

    def get_latest_profile_status(self, id: int):
        url = f'{self._catalog_api_base}/assets/{id}/profile'
        response = self._get(
            url=url
        )
        response['data']['client'] = self
        profile_response = ProfileRequest(**response['data'])
        return profile_response

    def cancel_profile(self, profile_req_id: int):
        url = f'{self._catalog_api_base}/assets/profile/{profile_req_id}/cancel'
        response = self._put(
            url=url,
            payload={}
        )
        return response['result']

    def get_asset_tags(self, id: int):
        if id is None:
            raise Exception('Asset id is required')
        url = f'{self._catalog_api_base}/assets/{id}/tags'
        response = self._get(
            url=url
        )
        tags = list(response['assetTags'])
        tags_list = []
        for res in tags:
            tags_list.append(AssetTag(**res))
        return tags_list

    def add_asset_tag(self, id: int, payload: dict):
        if id is None:
            raise Exception('Asset id is required')
        url = f'{self._catalog_api_base}/assets/{id}/tag'
        response = self._post(
            url=url, payload=payload
        )
        return AssetTag(**response)

    def add_asset_labels(self, id: int, payload: dict):
        if id is None:
            raise Exception('Asset id is required')
        url = f'{self._catalog_api_base}/assets/{id}/labels'
        response = self._put(
            url=url, payload=payload
        )
        labels = list(response)
        labels_list = []
        for lbl in labels:
            labels_list.append(AssetLabel(**lbl))
        return labels_list

    def get_asset_labels(self, id: int):
        if id is None:
            raise Exception('Asset id is required')
        url = f'{self._catalog_api_base}/assets/{id}/labels'
        response = self._get(
            url=url
        )
        labels = list(response)
        labels_list = []
        for lbl in labels:
            labels_list.append(AssetLabel(**lbl))
        return labels_list

    def get_analysis_pipeline(self, id: int) -> AnalyticsPipeline:
        response = self._get(
            url=f'{self._catalog_api_base}/analytics-pipelines/{id}'
        )
        return AnalyticsPipeline(**response['pipeline'])

    def check_connection(self, create_connection):
        create_conn = create_connection.__dict__
        config_prop = []
        for cp in create_connection.properties:
            config_prop.append(asdict(cp))
        create_conn['properties'] = config_prop
        payload = {
            'connection': create_conn
        }
        url = f'{self._catalog_api_base}/connections/check'
        response = self._post_form_data(
            url=url,
            payload=payload
        )
        check_res = ConnectionCheck(**response['result'])
        if check_res.status == ConnectionCheckStatus.SUCCESS.value:
            return check_res
        if check_res.status == ConnectionCheckStatus.FAILURE.value:
            raise TorchSdkException(f'Error: {check_res.message}')
        return check_res

    def create_connection(self, create_connection) -> Connection:
        create_conn = create_connection.__dict__
        config_prop = []
        # for cp in create_connection.properties:
        #     config_prop.append(asdict(cp))
        # create_conn['properties'] = config_prop
        payload = {
            'connection': create_conn
        }
        url = f'{self._catalog_api_base}/connections'
        response = self._post_form_data(
            url=url,
            payload=payload
        )
        return Connection(**response['connection'])

    def enable_rule(self, rule_id) -> Rule:
        payload = {
            "enable": True
        }
        response = self._put(
            url=f'{self._catalog_api_base}/rules/{rule_id}/enable',
            payload=payload
        )
        return Rule(**response['rule'])

    def disable_rule(self, rule_id) -> Rule:
        payload = {
            "enable": False
        }
        response = self._put(
            url=f'{self._catalog_api_base}/rules/{rule_id}/enable',
            payload=payload
        )
        return Rule(**response['rule'])

    def cancel_rule_execution(self, execution_id) -> RuleCancelResponse:
        response = self._put(
            url=f'{self._catalog_api_base}/rules/execution/{execution_id}/cancel',
            payload={}
        )
        return RuleCancelResponse(**response['result'])

    def execute_rule(self, rule_type: PolicyType, rule_id, sync=True, incremental=False,
                     failure_strategy: FailureStrategy = FailureStrategy.DoNotFail, pipeline_run_id=None,
                     policy_execution_request: PolicyExecutionRequest = None):
        """
        Description:
            To execute rule synchronously and asynchronously
        :param rule_type: (PolicyType) Type of rule to be executed
        :param rule_id: (String) id of the rule to be executed
        :param sync: (bool) optional Set it to False if asynchronous execution has to be done
        :param incremental: (bool) optional Set it to True if full execution has to be done
        :param failure_strategy: (enum) optional Set it to decide if it should fail at error,
            fail at warning or never fail
        :param policy_execution_request: (PolicyExecutionRequest) An optional parameter that allows you to provide
        additional options for executing the policy. It is an instance of the PolicyExecutionRequest modules class,
        which contains various properties that can be used to customize the policy execution, such as `executionType`,
        `markerConfigs`, `ruleItemSelections`, and more.
        """
        self.logger.info(
            f"Policy Execution Request: {policy_execution_request.to_dict() if policy_execution_request else 'None'}")

        rule_executor = Executor(rule_type, self, sync=sync)
        rule_executor.execute(rule_id, incremental, failure_strategy=failure_strategy, pipeline_run_id=pipeline_run_id,
                              policy_execution_request=policy_execution_request)
        return rule_executor

    def get_rule_status(self, rule_type, execution_id) -> RuleExecutionStatus:
        """
        Description:
            To get status of rule execution
        :param rule_type: (String) Type of rule to be executed
        :param execution_id: (String) ID of execution of the rule previously executed.
        """
        rule_executor = Executor(rule_type, self)
        return rule_executor.get_execution_status(execution_id)

    def get_rule_execution_result(self, rule_type, execution_id,
                                  failure_strategy: FailureStrategy = FailureStrategy.DoNotFail):
        """
        Description:
            To get result of rule execution
        :param rule_type: (String) Type of rule to be executed
        :param execution_id: (String) ID of execution of the rule previously executed.
        :param failure_strategy: (enum) optional Set it to decide if it should fail at error,
            fail at warning or never fail
        """
        rule_executor = Executor(rule_type, self)
        return rule_executor.get_execution_result(execution_id, failure_strategy=failure_strategy)

    def execute_dq_rule(self, rule_id, incremental=False, pipeline_run_id=None,
                        policy_execution_request: PolicyExecutionRequest = None) -> RuleExecution:

        self.logger.info(
            f"Invoked execute_dq_rule of the torch_http_client\n"
            f"Incremental flag: {incremental}\n"
            f"Policy Execution Request: {policy_execution_request.to_dict() if policy_execution_request else 'None'}"
        )
        payload_temp = dict()

        if policy_execution_request is None and not incremental:
            self.logger.info(
                "policy_execution_request is None and incremental is also false. Setting the executionType to FULL")
            payload_temp['executionType'] = "FULL"

        if pipeline_run_id is not None:
            self.logger.info(f"Setting up pipeline run ID: {pipeline_run_id}")
            payload_temp['pipelineRunId'] = pipeline_run_id

        if policy_execution_request is not None:
            self.logger.info(f"Marker config: {policy_execution_request.markerConfigs}")
            if policy_execution_request.executionType == ExecutionType.SELECTIVE and policy_execution_request.markerConfigs is None:
                raise TorchSdkException(
                    f'markerConfig is a mandatory parameter for execution type {ExecutionType.SELECTIVE}')
            else:
                self.logger.info("Converting policy_execution_request to dictionary")
                payload_temp = policy_execution_request.to_dict()

        elif incremental:
            self.logger.info("Setting the executionType to INCREMENTAL")
            payload_temp['executionType'] = "INCREMENTAL"

        self.logger.info(f"Executing DQ policy with the following payload: {payload_temp}")
        response = self._post(
            url=f'{self._catalog_api_base}/rules/data-quality/{rule_id}/executions',
            payload=payload_temp
        )
        return RuleExecution(**response['execution'])

    def execute_reconciliation_rule(self, rule_id, incremental=False, pipeline_run_id=None,
                                    policy_execution_request: PolicyExecutionRequest = None) -> RuleExecution:
        self.logger.info(
            f"Invoked execute_reconciliation_rule of the torch_http_client\n"
            f"Incremental flag: {incremental}\n"
            f"Policy Execution Request: {policy_execution_request.to_dict() if policy_execution_request else 'None'}"
        )
        payload_temp = dict()

        if policy_execution_request is None and not incremental:
            self.logger.info(
                "Policy execution request is None and incremental is also false. Setting the executionType to FULL")
            payload_temp['executionType'] = "FULL"

        if pipeline_run_id is not None:
            self.logger.info(f"Setting up pipeline run ID: {pipeline_run_id}")
            payload_temp['pipelineRunId'] = pipeline_run_id

        if policy_execution_request is not None:
            self.logger.info(f"Marker config: {policy_execution_request.markerConfigs}")
            if policy_execution_request.executionType == ExecutionType.SELECTIVE and policy_execution_request.markerConfigs is None:
                raise TorchSdkException(
                    f'markerConfig is a mandatory parameter for execution type {ExecutionType.SELECTIVE}')
            else:
                self.logger.info("Converting policy_execution_request to dictionary")
                payload_temp = policy_execution_request.to_dict()
                self.logger.info(f"Policy execution request to dictionary: {payload_temp}")
        elif incremental:
            self.logger.info("Setting the executionType to INCREMENTAL")
            payload_temp['executionType'] = "INCREMENTAL"

        self.logger.info(f"Executing Reconciliation policy with the following payload: {payload_temp}")
        response = self._post(
            url=f'{self._catalog_api_base}/rules/reconciliation/{rule_id}/executions',
            payload=payload_temp
        )
        return RuleExecution(**response['execution'])

    def get_dq_rule(self, rule_id) -> DataQualityRuleResource:
        response = self._get(
            url=f'{self._catalog_api_base}/rules/data-quality/{rule_id}'
        )
        response['client'] = self
        return DataQualityRuleResource(**response)

    def get_reconciliation_rule(self, rule_id) -> ReconciliationRuleResource:
        response = self._get(
            url=f'{self._catalog_api_base}/rules/reconciliation/{rule_id}'
        )
        response['client'] = self
        return ReconciliationRuleResource(**response)

    def get_policy(self, identifier) -> RuleResource:
        """
        Description:
                To get an existing policy from torch catalog
        :param identifier: uid or id of the policy
        :return:(Policy) policy class instance
        """

        torch_version = self.get_torch_version()
        actual_version = APIVersionUtils.extract_actual_version(torch_version)
        APIVersionUtils.validate_torch_version_for_rule_id_api(actual_version)

        if Version(MIN_TORCH_BACKEND_VERSION_FOR_RULE_ID_API) in SimpleSpec(f'<={actual_version}'):
            url = f'{self._catalog_api_base}/rules/{identifier}'
            response = self._get(url)
            response['client'] = self
            return Rule(**response)
        else:
            raise TorchSdkException(
                f'This API is supported starting version: {self.MIN_TORCH_BACKEND_VERSION_FOR_RULE_ID_API}')

    def get_dq_rule_execution_details(self, execution_id) -> ExecutionResult:
        response = self._get(
            url=f'{self._catalog_api_base}/rules/data-quality/executions/{execution_id}'
        )
        return ExecutionResult(**response)

    def get_reconciliation_rule_execution_details(self, execution_id) -> ExecutionResult:
        response = self._get(
            url=f'{self._catalog_api_base}/rules/reconciliation/executions/{execution_id}'
        )
        return ExecutionResult(**response)

    def get_reconciliation_rule_result(self, execution_id) -> RuleResult:
        response = self._get(
            url=f'{self._catalog_api_base}/rules/reconciliation/executions/{execution_id}/result'
        )
        return ReconciliationExecutionResult(**response)

    def get_dq_rule_result(self, execution_id) -> RuleResult:
        response = self._get(
            url=f'{self._catalog_api_base}/rules/data-quality/executions/{execution_id}/result'
        )
        return DataQualityExecutionResult(**response)

    def policy_executions(self, identifier, rule_type: RuleType, page=0, size=25,
                          sortBy='finishedAt:DESC') -> List[RuleExecutionSummary]:
        params = dict()
        params['page'] = page
        params['size'] = size
        params['sortBy'] = sortBy
        rule_id = identifier
        if not check_int(identifier):
            if rule_type.value == RuleType.DATA_QUALITY.value:
                rule_res = self.get_dq_rule(identifier)
            elif rule_type.value == RuleType.RECONCILIATION.value:
                rule_res = self.get_reconciliation_rule(identifier)
            rule_id = rule_res.rule.id

        response = self._get(
            url=f'{self._catalog_api_base}/rules/{rule_type.value.lower()}/{rule_id}/executions', params=params
        )
        executions = list(response['executions'])
        rule_execution_list = []
        for res in executions:
            execution = res['execution']
            rule_execution_list.append(RuleExecutionSummary(**execution))
        return rule_execution_list

    def get_all_rules(self, filter: PolicyFilter, page=0, size=25, withLatestExecution=True,
                      sortBy='updatedAt:DESC') -> RuleResult:
        params = dict()
        params['page'] = page
        params['size'] = size
        params['withLatestExecution'] = withLatestExecution
        params['sortBy'] = sortBy
        if filter.policyType is not None:
            params['ruleType'] = filter.policyType.name
            if filter.policyType.name == RuleType.RECONCILIATION.name:
                params['ruleType'] = params['ruleType'] + ',EQUALITY'
        if filter.lastExecutionResult is not None:
            params['resultStatus'] = filter.lastExecutionResult.name
        if filter.enable is not None:
            params['enabled'] = filter.enable
        if filter.active is not None and filter.active is True:
            params['ruleStatus'] = 'ACTIVE'
        else:
            params['ruleStatus'] = 'ALL'
        if filter.assets is not None:
            params['assetIds'] = ",".join([str(i.id) for i in filter.assets])
        if filter.data_sources is not None:
            params['assemblyIds'] = ",".join([str(i.id) for i in filter.data_sources])
        if filter.period is not None:
            finished_before_time, started_after_time = TimeRangeCalculator.calculate_time_range(filter)
            if started_after_time is not None:
                params['startedAfter'] = int(round(started_after_time.timestamp() * 1000))
            params['finishedBefore'] = int(round(finished_before_time.timestamp() * 1000))

        response = self._get(
            url=f'{self._catalog_api_base}/rules',
            params=params
        )
        rules = list(response['rules'])
        rule_list = []
        for res in rules:
            rule = res['rule']
            rule_list.append(DataQualityRule(**rule))
        return rule_list

    @staticmethod
    def now_ms():
        return int(round(time.time() * 1000))

    def _post(self, url, payload=None, as_json=True):
        now_ms = self.now_ms()
        if _HEADERS['accessKey'] is None:
            _HEADERS['accessKey'] = self.access_key
        if _HEADERS['secretKey'] is None:
            _HEADERS['secretKey'] = self.secret_key
        response = requests.post(
            url=url, headers=_HEADERS, json=payload, timeout=self._timeout, verify=False
        )
        self.logger.info(
            f" POST {url} "
            f"payload={json.dumps(payload)} "
            f"duration_ms={self.now_ms() - now_ms}"
        )

        return self._response(response, as_json)

    def _put(self, url, payload=None, as_json=True):
        now_ms = self.now_ms()
        if _HEADERS['accessKey'] is None:
            _HEADERS['accessKey'] = self.access_key
        if _HEADERS['secretKey'] is None:
            _HEADERS['secretKey'] = self.secret_key
        response = requests.put(
            url=url, headers=_HEADERS, json=payload, timeout=self._timeout, verify=False
        )
        self.logger.info(
            f" PUT {url} "
            f"payload={json.dumps(payload)} "
            f"duration_ms={self.now_ms() - now_ms}"
        )

        return self._response(response, as_json)

    def _get(self, url, params=None, as_json=True):
        now_ms = self.now_ms()
        if _HEADERS['accessKey'] is None:
            _HEADERS['accessKey'] = self.access_key
        if _HEADERS['secretKey'] is None:
            _HEADERS['secretKey'] = self.secret_key
        response = requests.get(
            url=url, params=params, headers=_HEADERS, timeout=self._timeout, verify=False
        )
        self.logger.info(
            f" GET {url} "
            f"duration_ms={self.now_ms() - now_ms}"
        )

        return self._response(response, as_json)

    def _post_form_data(self, url, payload=None, as_json=True):
        now_ms = self.now_ms()
        if _HEADERS['accessKey'] is None:
            _HEADERS['accessKey'] = self.access_key
        if _HEADERS['secretKey'] is None:
            _HEADERS['secretKey'] = self.secret_key

        params = {
            'request': json.dumps(payload)
        }
        data = MultipartEncoder(fields=params)
        _HEADERS['Content-Type'] = data.content_type
        response = requests.post(
            url=url, headers=_HEADERS, data=data, timeout=self._timeout, verify=False
        )
        self.logger.info(
            f" POST {url} "
            f"payload={json.dumps(payload)} "
            f"duration_ms={self.now_ms() - now_ms}"
        )

        return self._response(response, as_json)

    def _put_form_data(self, url, payload=None, as_json=True):
        now_ms = self.now_ms()
        if _HEADERS['accessKey'] is None:
            _HEADERS['accessKey'] = self.access_key
        if _HEADERS['secretKey'] is None:
            _HEADERS['secretKey'] = self.secret_key

        params = {
            'request': json.dumps(payload)
        }
        data = MultipartEncoder(fields=params)
        _HEADERS['Content-Type'] = data.content_type
        response = requests.put(
            url=url, headers=_HEADERS, data=data, timeout=self._timeout, verify=False
        )
        self.logger.info(
            f" POSPUTT {url} "
            f"payload={json.dumps(payload)} "
            f"duration_ms={self.now_ms() - now_ms}"
        )

        return self._response(response, as_json)

    def _delete(self, url, params=None, as_json=True):
        now_ms = self.now_ms()
        if _HEADERS['accessKey'] is None:
            _HEADERS['accessKey'] = self.access_key
        if _HEADERS['secretKey'] is None:
            _HEADERS['secretKey'] = self.secret_key
        response = requests.delete(
            url=url, params=params, headers=_HEADERS, timeout=self._timeout, verify=False
        )
        self.logger.info(
            f" DELETE {url} "
            f"duration_ms={self.now_ms() - now_ms}"
        )

        return response

    def _response(self, response, as_json):
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.logger.error(f' {response.text}')
            self._raise_api_error(e, response.text)

        return response.json() if as_json else response.text

    def _raise_api_error(self, e, text):
        raise APIError(text) from e



import os

# setup these 3 env vars in your environment.
# torch_credentials = {
#     'url': os.getenv('TORCH_CATALOG_URL', 'https://torch.acceldata.local:5443/torch'),
#     'access_key': os.getenv('TORCH_ACCESS_KEY', 'P04IM8FNQRUCRTU'),
#     'secret_key': os.getenv('TORCH_SECRET_KEY', 'E6LL9YUPMG4BDTJHT2VZD75HW0B8E5')
# }

torch_credentials = {
    'url': os.getenv('TORCH_CATALOG_URL', 'https://acceldata.dev.10.90.3.89.nip.io'),
    'access_key': os.getenv('TORCH_ACCESS_KEY', '21DK0CVJUAE0LET'),
    'secret_key': os.getenv('TORCH_SECRET_KEY', '1PT5E06TLCHKGELZCOEC7LLOCPYHJO')
}

ds_name = 'java_sdk_snowflake_ds'
ds_id = '2740'
asset_uid = 'java_sdk_snowflake_ds.FINANCE.FINANCE.CUSTOMERS_ACCOUNTS'
asset_id = 1202688
dq_policy_name = "selective_policy_dq"
recon_policy_name = "selective_recon_incremental"

# Profiling constants
table_asset_uid = "java_sdk_snowflake_ds.CUSTOMERS_DATABASE.CUSTOMERS.SELECTIVE_POLICY_DATA"
kafka_asset_uid = "sangeeta_kafka_ds.sangeeta_kaka_asset"
file_based_asset_uid = "sourav_gcs_data_source.gcs_incremental_file"

# Pipeline constants
# UID of the pipeline in torch
job_uid = "read_data_from_s3"
pipeline_uid = "torch.external.integration.demo-default"
# Name of the pipeline in torch
pipeline_name = "Flow ETL External Integration Demo - Default"
pipeline_uid_external = "torch.external.integration.demo-explicit.time"
# Name of the pipeline in torch
pipeline_name_external = "Flow ETL External Integration Demo - Explicit time"
# Name of the postgres datasource in torch
postgres_ds = "POSTGRES-DS"
# Name of the s3 datasource in torch
s3_ds = "S3-DS"
# Name of initial output asset which is the output of job copying generated csv data to s3 as csv
s3_customer = "s3_customers"
# Name of output asset which is the output of job joining data from 2nd csv and postgres by looking up department
# name for department id
s3_postgres_customers = "s3_customers_postgres"
# Name of postgres table where department id vs department name mapping has been created in demo schema
pg_table = "demo.dept_info"

pg_database = "customers"
# Name of postgres field to order sql query on
order_by = "customer_id"
# S3 bucket name
bucket = "airflow-sdk-demo"
# The location of initial csv file in s3 bucket
csv1 = 'demo/customers.csv'
# The location of csv file in s3 bucket after joining with data from postgres table
csv3 = 'demo/customers-with-department-info.csv'
# Number of entries to be generated
csv_size = 30

# DQ Policy Constants
dq_policy_backward_compatible_name = "selective_policy_dq_backward_compatible"
kafka_dq_policy_name = "kafka_dq_policy"
spark_sql_policy_name = "SelectiveDQPolicysparkSQLDynamicFilterVariable"
incremental_dq_policy_name = "selective_dq_incremental"
file_event_based_dq = "gcs_selective_policy"

# Recon Policy Constants
kafka_recon_policy_name = "kafka_recon_policy"
selective_recon_policy_backward_compatible_name = "selective_policy_recon_backward_compatible"  # backward compatible name
selective_recon_policy_name = "selective_policy_recon"
incremental_recon_policy_name = "selective_recon_incremental"
file_event_based_recon = "gcs_incremental_recon"

# Retry settings
RETRY_INTERVAL = 180  # 3 minutes in seconds
MAX_RETRIES = 6



from acceldata_sdk.models.ruleExecutionResult import RuleExecutionResult, RuleItemResult, RuleResult, RuleExecutionSummary
from enum import Enum
from dataclasses import dataclass
from typing import List
from acceldata_sdk.models.rule import (ShortSegment, Label, BackingAsset, RuleThresholdLevel, RuleTag, PolicyGroup, RuleResource, RuleExecution)
from acceldata_sdk.models.common_types import PolicyExecutionRequest

from acceldata_sdk.models.ruleExecutionResult import ExecutorConfig, SparkResourceConfig
from acceldata_sdk.constants import FailureStrategy, PolicyType

import logging

logger = logging.getLogger('dqrule')
logger.setLevel(logging.INFO)

class RootCauseAnalysis:
    def __init__(self, key, bad, good, badFraction, goodFraction, *args, **kwargs):
        self.key = key
        self.bad = bad
        self.good = good
        self.badFraction = badFraction
        self.goodFraction = goodFraction

    def __repr__(self):
        return f"RootCauseAnalysis({self.__dict__})"


class DataQualityRuleExecutionResult(RuleExecutionResult):
    def __init__(self, status, description=None, successCount=None, failureCount=None, rows=None, failedRows=None,
                 qualityScore=None, rca=None, *args, **kwargs):
        super().__init__(status, description, successCount, failureCount, qualityScore)
        self.failedRows = failedRows
        self.rows = rows
        if isinstance(rca, dict):
            self.rca = RootCauseAnalysis(**rca)
        else:
            self.rca = rca
    
    def __repr__(self):
        return f"DataQualityRuleExecutionResult({self.__dict__})"


class DataQualityRuleItemResult(RuleItemResult):
    def __init__(self, id, ruleItemId, threshold, weightage, isWarning=None, resultPercent=None, businessItemId=None,
                 success=None, error=None, *args, **kwargs):
        super().__init__(id, ruleItemId, threshold, weightage, isWarning, resultPercent, success, error)
        self.businessItemId = businessItemId

    def __repr__(self):
        return f"DataQualityRuleItemResult({self.__dict__})"


class DataQualityExecutionResult(RuleResult):
    def __init__(self, execution, items, meta=None, result=None, *args, **kwargs):
        if isinstance(execution, dict):
            self.execution = RuleExecutionSummary(**execution)
        else:
            self.execution = execution
        if isinstance(result, dict):
            self.result = DataQualityRuleExecutionResult(**result)
        else:
            self.result = result
        self.items = list()
        for obj in items:
            if isinstance(obj, dict):
                self.items.append(DataQualityRuleItemResult(**obj))
            else:
                self.items.append(obj)
        self.meta = meta
        self.executionId = self.execution.id

    def __repr__(self):
        return f"DataQualityExecutionResult({self.__dict__})"


class DataQualityMeasurementType(Enum):
    MISSING_VALUES = 'MISSING_VALUES'
    DATATYPE_MATCH = 'DATATYPE_MATCH'
    REGEX_MATCH = 'REGEX_MATCH'
    VALUES_IN_LIST = 'VALUES_IN_LIST'
    DISTINCTNESS_CHECK = 'DISTINCTNESS_CHECK'
    DUPLICATE_ROWS_CHECK = 'DUPLICATE_ROWS_CHECK'
    PRECISION_SCALE_CHECK = 'PRECISION_SCALE_CHECK'
    BUSINESS_MEASURE = 'BUSINESS_MEASURE'
    TAG_MATCH = 'TAG_MATCH'
    RANGE_MATCH = 'RANGE_MATCH'
    SIZE_CHECK = 'SIZE_CHECK'
    CUSTOM = 'CUSTOM'
    UDF_PREDICATE = 'UDF_PREDICATE'


@dataclass
class Item:
    id = None
    ruleId = None
    measurementType = None
    executionOrder = None
    columnName = None
    value = None
    ruleExpression = None
    resultThreshold = None
    weightage = None
    ruleVersion = None
    businessExplanation = None
    labels = None
    isWarning = None

    def __init__(self,
                 id=None,
                 ruleId=None,
                 measurementType=None,
                 executionOrder=None,
                 columnName=None,
                 value=None,
                 ruleExpression=None,
                 resultThreshold=None,
                 weightage=None,
                 ruleVersion=None,
                 businessExplanation=None,
                 labels=None,
                 isWarning=None,
                 *args, **kwargs):
        self.ruleId = ruleId
        self.id = id
        if isinstance(measurementType, dict):
            self.measurementType = DataQualityMeasurementType(**measurementType)
        else:
            self.measurementType = measurementType
        self.executionOrder = executionOrder
        self.columnName = columnName
        self.value = value
        self.ruleExpression = ruleExpression
        self.resultThreshold = resultThreshold
        self.weightage = weightage
        self.ruleVersion = ruleVersion
        self.businessExplanation = businessExplanation
        self.labels = list()
        for obj in labels:
            if isinstance(obj, dict):
                self.labels.append(Label(**obj))
            else:
                self.labels.append(obj)
        self.isWarning = isWarning

    def __repr__(self):
        return f"Item({self.__dict__})"


class DataQualityRuleDetails:
    def __init__(self, ruleId, backingAssetId,
                 items, isSegmented, segments, id=None, *args, **kwargs):
        self.id = id
        self.ruleId = ruleId
        self.backingAssetId = backingAssetId
        self.items = list()
        for obj in items:
            if isinstance(obj, dict):
                self.items.append(Item(**obj))
            else:
                self.items.append(obj)
        self.isSegmented = isSegmented
        self.segments = list()
        if segments is not None:
            for obj in segments:
                if isinstance(obj, dict):
                    self.segments.append(ShortSegment(**obj))
                else:
                    self.segments.append(obj)
        self.columnMappings = list()

    def __repr__(self):
        return f"DataQualityRuleDetails({self.__dict__})"


class ChannelType(Enum):
    JIRA = 'JIRA'
    EMAIL = 'EMAIL'
    SLACK = 'SLACK'
    HANGOUT = 'HANGOUT'
    WEBHOOK = 'WEBHOOK'


class NotifyOn(Enum):
    ERROR = 'ERROR'
    WARNING = 'WARNING'
    SUCCESS = 'SUCCESS'
    ALL = 'ALL'


@dataclass
class NotificationPayload:
    configuredList: List
    notifyOn: List[NotifyOn]
    tags: List[str] = None


class DataQualityRule:
    def __init__(self, name=None, description=None, type=None, enabled=None, schedule=None, scheduled=None,
                 notificationChannels=None, backingAssets=None, createdAt=None, updatedAt=None, thresholdLevel=None,
                 archived=None, archivalReason=None, tenantId=None, createdBy=None, lastUpdatedBy=None, tags=None,
                 segments=None, version=None, executorConfig=None, labels=None, isProtectedResource=None,
                 policyGroups=None, id=None, sparkResourceConfig=None, *args, **kwargs):
        self.id = id
        self.name = name
        self.description = description
        self.type = type
        self.enabled = enabled
        self.schedule = schedule
        self.scheduled = scheduled
        self.notificationChannels = notificationChannels
        self.backingAssets = list()
        if backingAssets is not None:
            for obj in backingAssets:
                if isinstance(obj, dict):
                    self.backingAssets.append(BackingAsset(**obj))
                else:
                    self.backingAssets.append(obj)
        self.createdAt = createdAt
        self.updatedAt = updatedAt
        if isinstance(thresholdLevel, dict):
            self.thresholdLevel = RuleThresholdLevel(**thresholdLevel)
        else:
            self.thresholdLevel = thresholdLevel
        self.archived = archived
        self.archivalReason = archivalReason
        self.tenantId = tenantId
        self.createdBy = createdBy
        self.lastUpdatedBy = lastUpdatedBy

        self.tags = list()
        if tags is not None:
            for obj in tags:
                if isinstance(obj, dict):
                    self.tags.append(RuleTag(**obj))
                else:
                    self.tags.append(obj)
        self.segments = list()
        if segments is not None:
            for obj in segments:
                if isinstance(obj, dict):
                    self.segments.append(ShortSegment(**obj))
                else:
                    self.segments.append(obj)
        self.version = version
        if isinstance(executorConfig, dict):
            self.executorConfig = ExecutorConfig(**executorConfig)
        else:
            self.executorConfig = executorConfig
        if isinstance(sparkResourceConfig, dict):
            self.sparkResourceConfig = SparkResourceConfig(**sparkResourceConfig)
        else:
            self.sparkResourceConfig = sparkResourceConfig

        self.labels = list()
        if labels is not None:
            for obj in labels:
                if isinstance(obj, dict):
                    self.labels.append(Label(**obj))
                else:
                    self.labels.append(obj)
        self.isProtectedResource = isProtectedResource

        self.policyGroups = list()
        if policyGroups is not None:
            for obj in policyGroups:
                if isinstance(obj, dict):
                    self.policyGroups.append(PolicyGroup(**obj))
                else:
                    self.policyGroups.append(obj)

    def __repr__(self):
        return f"DataQualityRule({self.__dict__})"


class DataQualityRuleResource(RuleResource):
    def __init__(self, rule, details, client=None, *args, **kwargs):
        if isinstance(rule, dict):
            self.rule = DataQualityRule(**rule)
        else:
            self.rule = rule
        if isinstance(details, dict):
            self.details = DataQualityRuleDetails(**details)
        else:
            self.details = details
        self.client = client

    def execute(self, sync=True, incremental=False, failure_strategy: FailureStrategy = FailureStrategy.DoNotFail,
                policy_execution_request: PolicyExecutionRequest = None) -> RuleExecution:
        logger.info(
            f"Policy Execution Request: {policy_execution_request.to_dict() if policy_execution_request else 'None'}")
        return self.client.execute_rule(PolicyType.DATA_QUALITY, self.rule.id, sync, incremental, failure_strategy,
                                        policy_execution_request)

    def get_executions(self, page=0, size=25, sortBy='finishedAt:DESC'):
        return self.client.policy_executions(self.rule.id, PolicyType.DATA_QUALITY, page, size, sortBy)

    def __repr__(self):
        return f"DataQualityRuleResource({self.__dict__})"



from enum import Enum, auto
from typing import List, Optional, Dict, Any


class ExecutionType(Enum):
    SELECTIVE = 'SELECTIVE'
    FULL = 'FULL'
    INCREMENTAL = 'INCREMENTAL'

    def to_dict(self):
        return self.name


class MarkerConfig:
    def __init__(self, type: str):
        self.type = type

    def to_dict(self) -> Dict:
        return {'type': self.type}


class BoundsIdMarkerConfig(MarkerConfig):
    def __init__(self, baseAssetId: Optional[int] = None, idColumnName: str = "",
                 fromId: Optional[int] = None, toId: Optional[int] = None):
        """
        Initializes a BoundsIdMarkerConfig object.

        Args:
            baseAssetId (Optional[int]): The base asset ID. Defaults to None.
            idColumnName (str): The ID column name. Defaults to "".
            fromId (Optional[int]): The starting ID. Defaults to None.
            toId (Optional[int]): The ending ID. Defaults to None.
        """
        super().__init__("bound")
        self.baseAssetId = baseAssetId
        self.idColumnName = idColumnName
        self.fromId = fromId
        self.toId = toId

    def to_dict(self) -> Dict:
        data = super().to_dict()
        data.update({
            'baseAssetId': self.baseAssetId,
            'idColumnName': self.idColumnName,
            'fromId': self.fromId,
            'toId': self.toId
        })
        return data


class BoundsFileEventMarkerConfig(MarkerConfig):
    def __init__(self, baseAssetId: Optional[int] = None, dateColumnName: str = "event_time",
                 timeZoneId: str = "UTC", fromDate: Optional[str] = None, toDate: Optional[str] = None):
        """
        Initializes a BoundsFileEventMarkerConfig object.

        Args:
            baseAssetId (Optional[int]): The base asset ID. Defaults to None.
            dateColumnName (str): The name of the date column. Defaults to "event_time".
            timeZoneId (str): The ID of the time zone. Defaults to "UTC".
            fromDate (Optional[str]): The starting date. Defaults to None.
            toDate (Optional[str]): The ending date. Defaults to None.
        """
        super().__init__("boundFileEvent")
        self.baseAssetId = baseAssetId
        self.dateColumnName = dateColumnName
        self.timeZoneId = timeZoneId
        self.fromDate = fromDate
        self.toDate = toDate

    def to_dict(self) -> Dict:
        data = super().to_dict()
        data.update({
            'baseAssetId': self.baseAssetId,
            'dateColumnName': self.dateColumnName,
            'timeZoneId': self.timeZoneId,
            'fromDate': self.fromDate,
            'toDate': self.toDate
        })
        return data


class BoundsDateTimeMarkerConfig(MarkerConfig):
    def __init__(self, baseAssetId: Optional[int] = None, dateColumnName: str = "",
                 format: str = "", timeZoneId: str = "UTC", fromDate: Optional[str] = None,
                 toDate: Optional[str] = None):
        """
        Initializes a BoundsDateTimeMarkerConfig object.

        Args:
            baseAssetId (Optional[int]): The base asset ID. Defaults to None.
            dateColumnName (str): The name of the date column. Defaults to "".
            format (str): The format of the date. Defaults to "".
            timeZoneId (str): The ID of the time zone. Defaults to "UTC".
            fromDate (Optional[str]): The starting date. Defaults to None.
            toDate (Optional[str]): The ending date. Defaults to None.
        """
        super().__init__("boundDateTime")
        self.baseAssetId = baseAssetId
        self.dateColumnName = dateColumnName
        self.format = format
        self.timeZoneId = timeZoneId
        self.fromDate = fromDate
        self.toDate = toDate

    def to_dict(self) -> Dict:
        data = super().to_dict()
        data.update({
            'baseAssetId': self.baseAssetId,
            'dateColumnName': self.dateColumnName,
            'format': self.format,
            'timeZoneId': self.timeZoneId,
            'fromDate': self.fromDate,
            'toDate': self.toDate
        })
        return data


class TimestampBasedMarkerConfig(MarkerConfig):
    def __init__(self, baseAssetId: Optional[int] = None, format: str = "",
                 timeZoneId: str = "UTC", initialOffset: Optional[str] = None,
                 toDateTime: Optional[str] = None, fromOffsets: Optional[str] = None):
        """
        Initializes a TimestampBasedMarkerConfig object.

        Args:
            baseAssetId (Optional[int]): The base asset ID. Defaults to None.
            format (str): The format of the timestamp. Defaults to "".
            timeZoneId (str): The ID of the time zone. Defaults to "UTC".
            initialOffset (Optional[str]): The initial offset of the timestamp. Defaults to None.
            toDateTime (Optional[str]): The end date and time of the timestamp. Defaults to None.
            fromOffsets (Optional[str]): The offsets of the timestamp. Defaults to None.
        """
        super().__init__("timestamp")
        self.baseAssetId = baseAssetId
        self.format = format
        self.timeZoneId = timeZoneId
        self.initialOffset = initialOffset
        self.toDateTime = toDateTime
        self.fromOffsets = fromOffsets

    def to_dict(self) -> Dict:
        data = super().to_dict()
        data.update({
            'baseAssetId': self.baseAssetId,
            'format': self.format,
            'timeZoneId': self.timeZoneId,
            'initialOffset': self.initialOffset,
            'toDateTime': self.toDateTime,
            'fromOffsets': self.fromOffsets
        })
        return data


class AssetMarkerConfig:
    def __init__(self, assetId: Optional[int] = None, markerConfig: MarkerConfig = None):
        self.assetId = assetId
        self.markerConfig = markerConfig

    def to_dict(self):
        return {
            'assetId': self.assetId,
            'markerConfig': self.markerConfig.to_dict()
        }


class YunikornConfig:
    def __init__(self, minExecutors: int, maxExecutors: int, executorCores: int, executorMemory: str, driverCores: int,
                 driverMemory: str):
        """
        Initializes an instance of the YunikornConfig class.

        Args:
            minExecutors (int): The minimum number of executors.
            maxExecutors (int): The maximum number of executors.
            executorCores (int): The number of cores per executor.
            executorMemory (str): The memory configuration for the executors.
            driverCores (int): The number of cores for the driver.
            driverMemory (str): The memory configuration for the driver.
        """
        self.minExecutors = minExecutors
        self.maxExecutors = maxExecutors
        self.executorCores = executorCores
        self.executorMemory = executorMemory
        self.driverCores = driverCores
        self.driverMemory = driverMemory

    def to_dict(self):
        return self.__dict__


class SparkResourceConfig:
    def __init__(self, yunikorn: YunikornConfig, additionalConfiguration: Dict[str, str]):
        """
        Initializes an instance of the SparkResourceConfig class.

        Args:
            yunikorn (YunikornConfig): The configuration for Yunikorn.
            additionalConfiguration (Dict[str, str]): Additional configuration for Spark.
        """
        self.yunikorn = yunikorn
        self.additionalConfiguration = additionalConfiguration

    def to_dict(self):
        return {
            'yunikorn': self.yunikorn.to_dict(),
            'additionalConfiguration': self.additionalConfiguration
        }


class Mapping:
    def __init__(self, key: str, isColumnVariable: Optional[bool] = None, value: Optional[str] = None):
        """
        Initializes a Mapping object.

        Args:
            key (str): The key of the mapping.
            isColumnVariable (Optional[bool]): A flag indicating if the mapping is a column variable.
            value (Optional[str]): The value of the mapping.
        """
        self.key = key
        self.isColumnVariable = isColumnVariable
        self.value = value

    def to_dict(self) -> Dict[str, Any]:
        return {
            "key": self.key,
            "isColumnVariable": self.isColumnVariable,
            "value": self.value
        }


class RuleSparkSQLDynamicFilterVariableMapping:
    def __init__(self, ruleName: str = None, mapping: Optional[List[Mapping]] = None):
        """
        Initializes a `RuleSparkSQLDynamicFilterVariableMapping` object.
        Args:
            ruleName (str, optional): The name of the rule. Defaults to None.
            mapping (List[Mapping], optional): The list of mappings. Defaults to None.
        """
        self.ruleName = ruleName
        self.mapping = mapping if mapping is not None else []

    def to_dict(self) -> Dict[str, Any]:
        return {
            "ruleName": self.ruleName,
            "mapping": [m.to_dict() for m in self.mapping]
        }


class PolicyExecutionRequest:
    def __init__(self, executionType: ExecutionType,
                 markerConfigs: Optional[List[AssetMarkerConfig]] = None,
                 ruleItemSelections: Optional[List[int]] = None,
                 includeInQualityScore: bool = True,
                 pipelineRunId: Optional[int] = None,
                 sparkSQLDynamicFilterVariableMapping: Optional[List[RuleSparkSQLDynamicFilterVariableMapping]] = None,
                 sparkFilterSelectedColumns: Optional[List[str]] = None,
                 sparkResourceConfig: Optional[SparkResourceConfig] = None):
        """
        Initializes a PolicyExecutionRequest object.

        Args:
            executionType (ExecutionType): The type of execution. Supported types: (SELECTIVE, FULL, INCREMENTAL)
            markerConfigs (List[AssetMarkerConfig], optional): The list of marker configurations. Defaults to None.
            ruleItemSelections (List[int], optional): The list of rule item selections. Defaults to None. If not passed all the rule item definitions will be executed.
            includeInQualityScore (bool, optional): Whether to include in quality score. Defaults to True.
            pipelineRunId (int, optional): ID of the pipeline run for attaching policy execution to the asset. Defaults to None.
            sparkSQLDynamicFilterVariableMapping (List[RuleSparkSQLDynamicFilterVariableMapping], optional): The list of Spark SQL dynamic filter variable mappings applicable for a Data Quality Policy. Defaults to None.
            sparkFilterSelectedColumns (List[str], optional): The list of Spark filter selected columns. Defaults to None.
            sparkResourceConfig (SparkResourceConfig, optional): The Spark resource configuration. Defaults to None.
        """
        self.executionType = executionType
        self.markerConfigs = markerConfigs
        self.ruleItemSelections = ruleItemSelections
        self.includeInQualityScore = includeInQualityScore
        self.pipelineRunId = pipelineRunId
        self.sparkSQLDynamicFilterVariableMapping = sparkSQLDynamicFilterVariableMapping if sparkSQLDynamicFilterVariableMapping is not None else []
        self.sparkFilterSelectedColumns = sparkFilterSelectedColumns
        self.sparkResourceConfig = sparkResourceConfig

    def to_dict(self) -> Dict[str, Any]:
        return {
            key: value for key, value in {
                "executionType": self.executionType.to_dict(),
                "markerConfigs": [config.to_dict() for config in self.markerConfigs] if self.markerConfigs else None,
                "ruleItemSelections": self.ruleItemSelections,
                "includeInQualityScore": self.includeInQualityScore,
                "pipelineRunId": self.pipelineRunId,
                "sparkSQLDynamicFilterVariableMapping": [mapping.to_dict() for mapping in
                                                         self.sparkSQLDynamicFilterVariableMapping],
                "sparkFilterSelectedColumns": self.sparkFilterSelectedColumns,
                "sparkResourceConfig": self.sparkResourceConfig.to_dict() if self.sparkResourceConfig else None
            }.items() if value is not None
        }



from datetime import datetime, timedelta


class TimeRangeCalculator:

    def __init__(self):
        pass

    @staticmethod
    def check_int(input_str):
        try:
            int(input_str)
            return True
        except ValueError:
            return False

    @staticmethod
    def calculate_minutes_before_time(minutes):
        """
        Calculate the time range starting before the given number of minutes from the current time.
        :param minutes: Number of minutes to subtract from the current time.
        :return: A tuple containing two datetime objects -
                 the starting time before the given minutes and
                 the ending time as the current datetime.
        """
        now = datetime.now()
        finished_before_time = now
        started_after_time = finished_before_time - timedelta(minutes=minutes)
        return started_after_time, finished_before_time

    @staticmethod
    def calculate_hours_before_time(hours):
        """
        Calculate the time range starting before the given number of hours from the current time.
        :param hours: Number of hours to subtract from the current time.
        :return: A tuple containing two datetime objects -
                 the starting time before the given hours and
                 the ending time as the current datetime.
        """
        finished_before_time = datetime.now()
        started_after_time = finished_before_time - timedelta(hours=hours)
        return started_after_time, finished_before_time

    @staticmethod
    def calculate_days_before_time(days):
        """
        Calculate the time range starting before the given number of days from today's midnight.
        :param days: Number of days to subtract from today's midnight.
        :return: A tuple containing two datetime objects -
                 the starting time before the given days from the beginning of the current day and
                 the ending time as the beginning of the current day.
        """
        now = datetime.now()
        min_time = datetime.min.time()
        started_after_time = datetime.combine((now - timedelta(days=days)).date(), min_time)
        finished_before_time = datetime.combine(now.date(), min_time)
        return started_after_time, finished_before_time

    @staticmethod
    def calculate_time_range(filter):
        if filter.period is not None:
            now = datetime.now()
            finished_before_time = now
            started_after_time = None  # Initialize to None
            if filter.period == ExecutionPeriod.Last15minutes:
                started_after_time, finished_before_time = calculate_minutes_before_time(15)
            elif filter.period == ExecutionPeriod.Last30minutes:
                started_after_time, finished_before_time = calculate_minutes_before_time(30)
            elif filter.period == ExecutionPeriod.Last1hour:
                started_after_time, finished_before_time = calculate_hours_before_time(1)
            elif filter.period == ExecutionPeriod.Last3hours:
                started_after_time, finished_before_time = calculate_hours_before_time(3)
            elif filter.period == ExecutionPeriod.Last6hours:
                started_after_time, finished_before_time = calculate_hours_before_time(6)
            elif filter.period == ExecutionPeriod.Last12hours:
                started_after_time, finished_before_time = calculate_hours_before_time(12)
            elif filter.period == ExecutionPeriod.Last24hours:
                started_after_time, finished_before_time = calculate_hours_before_time(24)
            elif filter.period == ExecutionPeriod.Today:
                started_after_time = datetime.combine(datetime.now().date(), datetime.min.time())
            elif filter.period == ExecutionPeriod.Yesterday:
                started_after_time, finished_before_time = calculate_days_before_time(1)
            elif filter.period == ExecutionPeriod.Last7days:
                started_after_time, finished_before_time = calculate_days_before_time(7)
            elif filter.period == ExecutionPeriod.Thismonth:
                started_after_time = datetime(datetime.now().year, datetime.now().month, 1)
            elif filter.period == ExecutionPeriod.Last1month:
                started_after_time, finished_before_time = calculate_months_before_time(1)
            elif filter.period == ExecutionPeriod.Last3month:
                started_after_time, finished_before_time = calculate_months_before_time(3)
            return finished_before_time, started_after_time



import unittest
import random
from unittest.mock import Mock, patch, call
from acceldata_sdk.constants import SdkSupportedVersions, TorchBuildVersion
from acceldata_sdk.torch_client import TorchClient
from acceldata_sdk.models.pipeline import CreatePipeline, PipelineMetadata, PipelineRunResult, PipelineRunStatus
import pprint
import time
import os
from datetime import datetime
from acceldata_sdk.models.job import CreateJob, JobMetadata, Node
from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_sdk.events.log_events import LogEvent
import acceldata_sdk.constants as const
from acceldata_sdk.models.ruleExecutionResult import RuleType, PolicyFilter, ExecutionPeriod
from acceldata_sdk.models.tags import AssetLabel, CustomAssetMetadata
from acceldata_sdk.models.profile import Profile, ProfileRequest, ProfilingType

pp = pprint.PrettyPrinter(indent=4)

# Pipeline constants
pipeline_uid = "adoc.test.pipeline.sanity"
pipeline_uid_with_continuation_id = "adoc.test.pipeline.sanity.continuation"
pipeline_uid_with_continuation_id_name = "ADOC Test Pipeline Sanity ContinuationId"
pipeline_name = "ADOC Test Pipeline Sanity"
job_uid_customers_read = "customers.read"
job_uid_generate_sales = "customers.generate_sales"
job_uid_sales = "aggregated_sales"

# DQ Policy constants
dq_policy_name = "snowflake_dq_automation_null_check_id_success"

# Reconciliation Policy constants
recon_policy_name = "Athena_automation_Hashed_Data_Equality_4_success"

# Datasource name
datasource_name = "Snowflake_stg"

# Asset UID
asset_uid_kafka = "Sample_Kafka.member_updates"
asset_uid_file = "S3_Demo.trans_detail"
asset_uid_table = "Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.CALL_CENTER"


class TestTorchClientIntegration(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Set up common variables or configurations if needed
        pass

    def setUp(self):
        # Create an instance of TorchClient with mock values for testing
        try:
            self.mock_url = os.getenv('TORCH_CATALOG_URL')
            self.mock_access_key = os.getenv('TORCH_ACCESS_KEY')
            self.mock_secret_key = os.getenv('TORCH_SECRET_KEY')
            self.torch_client = TorchClient(url=self.mock_url, access_key=self.mock_access_key,
                                            secret_key=self.mock_secret_key)
        except KeyError as e:
            raise ValueError(f"Environment variable {e} is not set. Please set all required variables.")

    def tearDown(self):
        pass

    # Pipeline test cases

    def test_get_torch_version(self):
        result = self.torch_client.get_torch_version()
        # Assertions
        pp.pprint("Validating get_torch_version")
        pp.pprint(result)
        self.assertIsNotNone(result)

    def test_get_supported_sdk_versions(self):
        result = self.torch_client.get_supported_sdk_versions()
        # Assertions
        pp.pprint("Validating get_supported_sdk_versions")
        pp.pprint(result)
        self.assertIsNotNone(result)

    def test_create_pipeline(self):
        meta = PipelineMetadata(owner='sdk/pipeline-user', team='TORCH', codeLocation='...')
        create_pipeline_payload = CreatePipeline(
            uid=pipeline_uid,
            name=pipeline_name,
            description=f'The has been created from torch-sdk',
            meta=meta,
            context={'pipeline_uid': pipeline_uid, 'pipeline_name': pipeline_name}
        )
        pp.pprint("Validating create_pipeline")
        create_pipeline_response = self.torch_client.create_pipeline(pipeline=create_pipeline_payload)
        pp.pprint(create_pipeline_response)
        pp.pprint("Validating get_pipeline by id")
        pipeline = self.torch_client.get_pipeline(create_pipeline_response.id)
        pp.pprint(pipeline)
        assert pipeline is not None

    def test_get_pipeline_by_uid(self):
        pp.pprint("Validating get_pipeline by UID")
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        pp.pprint(pipeline)
        assert pipeline is not None

    def test_create_pipeline_run(self):
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        pp.pprint("Validating create_pipeline_run")
        pipeline_run = pipeline.create_pipeline_run()
        pp.pprint(pipeline_run)
        run_with_id = pipeline.get_run(pipeline_run.id)
        pp.pprint(run_with_id)
        assert pipeline_run == run_with_id

    def test_create_pipeline_run_with_continuation_id(self):
        create_pipeline_payload = CreatePipeline(
            uid=pipeline_uid_with_continuation_id,
            name=pipeline_uid_with_continuation_id_name,
            description=f'The has been created from torch-sdk',
        )
        pipeline = self.torch_client.create_pipeline(pipeline=create_pipeline_payload)
        pp.pprint("Validating test_create_pipeline_run_with_continuation_id")
        pipeline_run = pipeline.create_pipeline_run(continuation_id=f'run.{random.random()}')
        pp.pprint(pipeline_run)
        run_with_id = pipeline.get_run(pipeline_run.id)
        pp.pprint(run_with_id)
        assert pipeline_run == run_with_id

    def test_get_pipelines(self):
        pp.pprint("Validating get_pipelines")
        pipelines = self.torch_client.get_pipelines()
        pp.pprint(pipelines)
        assert pipelines is not None

    def test_get_pipeline_runs(self):
        pp.pprint("Validating test_get_pipeline_runs")
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        pipeline_runs = pipeline.get_runs()
        pp.pprint(pipeline_runs)
        assert pipeline_runs is not None

    def test_latest_pipeline_run(self):
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        pp.pprint("Validating get_latest_pipeline_run")
        latest_run = pipeline.get_latest_pipeline_run()
        pp.pprint(latest_run)
        pp.pprint("Validating get_runs")
        pipeline_runs = pipeline.get_runs()
        pp.pprint(pipeline_runs)
        assert latest_run == pipeline_runs[0]

    def test_create_root_span(self):
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        latest_run = pipeline.get_latest_pipeline_run()
        pp.pprint("Create root span validation")
        root_span = latest_run.create_span(uid=f'{pipeline_uid}.span')
        pp.pprint(root_span)
        assert root_span is not None

    def test_create_job_for_pipeline(self):
        inputs = [Node(asset_uid=f'S3-DS.s3_customers')]
        outputs = [Node(job_uid=job_uid_generate_sales)]
        context_job = {'job': 'data_gene', 'time': str(datetime.now()), 'uid': 'customers.read',
                       'operator': 'write_file_func'}
        metadata = JobMetadata('Jason', 'COKE', 'https://github.com/coke/reports/customers.kt')
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        latest_run = pipeline.get_latest_pipeline_run()
        pp.pprint("Creating Job using pipeline validation")
        job = CreateJob(
            uid=job_uid_customers_read,
            name=f'{job_uid_customers_read} Job',
            pipeline_run_id=latest_run.id,
            description=f'{job_uid_customers_read} created using torch SDK',
            inputs=inputs,
            outputs=outputs,
            meta=metadata,
            context=context_job,
            bounded_by_span=True,
            span_uid=f'{job_uid_customers_read}.span'
        )
        job = pipeline.create_job(job)
        job_span = latest_run.get_span(f'{job_uid_customers_read}.span')
        pp.pprint(job_span)
        job_span.end()
        assert job is not None

    def test_create_job_for_pipeline_run(self):
        inputs = [Node(job_uid=job_uid_customers_read)]
        outputs = [Node(asset_uid=f'S3-DS.orders')]
        context_job = {'job': 'data_gene', 'time': str(datetime.now()), 'uid': 'customers.data-generation',
                       'operator': 'write_file_func'}
        metadata = JobMetadata('Jason', 'COKE', 'https://github.com/coke/reports/customers.kt')
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        latest_run = pipeline.get_latest_pipeline_run()
        pp.pprint("Creating Job using pipeline_run validation")
        job = CreateJob(
            uid=job_uid_generate_sales,
            name=f'{job_uid_generate_sales} Job',
            description=f'{job_uid_generate_sales} created using torch SDK',
            inputs=inputs,
            outputs=outputs,
            meta=metadata,
            context=context_job,
            bounded_by_span=True,
            span_uid=f'{job_uid_generate_sales}.span'
        )
        job = latest_run.create_job(job)
        job_span = latest_run.get_span(f'{job_uid_generate_sales}.span')
        pp.pprint(job_span)
        job_span.end()
        assert job is not None

    def test_create_child_span(self):
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        latest_run = pipeline.get_latest_pipeline_run()
        sales_span = latest_run.get_span(f'{job_uid_generate_sales}.span')
        pp.pprint("Validating create child span")
        span_child = sales_span.create_child_span(
            uid=f'{job_uid_sales}.span',
            context_data={
                'time': str(datetime.now())
            },
            associatedJobUids=[job_uid_sales])
        pp.pprint(span_child)
        pp.pprint("Validating span events")
        span_child.start()
        span_child.send_event(GenericEvent(
            context_data={'client_time': str(datetime.now()), 'total_file': 1,
                          'schema': 'name,address,dept_id'},
            event_uid="customers.data.generation.metadata"))
        span_child.send_event(LogEvent(
            context_data={'Size': 100, 'total_file': 1,
                          'schema': 'name,address,dept_id'},
            log_data="Customer data generated successfully."))
        span_child.end()
        span_child2 = sales_span.create_child_span(
            uid=f'{job_uid_sales}2.span',
            context_data={
                'time': str(datetime.now())
            },
            associatedJobUids=[job_uid_sales])
        pp.pprint(span_child2)
        pp.pprint("Validating span events")
        span_child2.start()
        span_child2.abort()
        sales_span.failed()
        if sales_span.is_root():
            query_span_id = latest_run.get_span(span_child.span.id)
            assert query_span_id is not None
            pp.pprint("query_span_id")
            pp.pprint(query_span_id)
            child_spans = sales_span.get_child_spans()
            assert child_spans is not None
            pp.pprint("Child spans")
            pp.pprint(child_spans)

    def test_spans_pipeline_run(self):
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        latest_run = pipeline.get_latest_pipeline_run()
        spans = latest_run.get_spans()
        assert spans is not None

    def test_details_pipeline_run(self):
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        latest_run = pipeline.get_latest_pipeline_run()
        details = latest_run.get_details()
        assert details is not None

    def test_update_pipeline_run(self):
        pipeline = self.torch_client.get_pipeline(pipeline_uid)
        latest_run = pipeline.get_latest_pipeline_run()
        update_pipeline_run_res = latest_run.update_pipeline_run(
            context_data={'key1': 'value2', 'name': 'backend'},
            result=PipelineRunResult.SUCCESS,
            status=PipelineRunStatus.COMPLETED
        )
        assert update_pipeline_run_res is not None

    def test_delete_pipeline(self):
        pipeline = self.torch_client.get_pipeline(pipeline_uid_with_continuation_id)
        delete_res = pipeline.delete()
        assert delete_res is not None

    # DQ Policy Test Cases
    def test_get_dq_policy(self):
        pp.pprint("Validating test test_get_dq_policy")
        dq_rule = self.torch_client.get_policy(const.PolicyType.DATA_QUALITY, dq_policy_name)
        pp.pprint(dq_rule)
        assert dq_rule is not None

    def test_get_dq_policy_without_type(self):
        pp.pprint("Validating test test_get_dq_policy_without_type")
        dq_rule = self.torch_client.get_policy(identifier=dq_policy_name)
        pp.pprint(dq_rule)
        assert dq_rule is not None

    def test_get_all_dq_policy(self):
        pp.pprint("Validating test test_get_all_dq_policy")
        policy_filter = PolicyFilter(policyType=RuleType.DATA_QUALITY, enable=True)
        dq_rules = self.torch_client.list_all_policies(filter=policy_filter)
        pp.pprint(dq_rules)
        assert dq_rules is not None

    def test_cancel_dq_policy(self):
        pp.pprint("Validating test test_cancel_dq_policy")
        dq_rule = self.torch_client.get_policy(const.PolicyType.DATA_QUALITY, dq_policy_name)
        async_executor = dq_rule.execute(sync=False)
        pp.pprint("Validating test async_executor")
        pp.pprint(async_executor)
        if async_executor.errorMessage is None:
            async_execution_status = async_executor.get_status()
            pp.pprint('async_execution_status')
            pp.pprint(async_execution_status)
            async_execution_cancel = async_executor.cancel()
            pp.pprint('async_execution_cancel')
            pp.pprint(async_execution_cancel)

    def test_execute_dq_policy(self):
        pp.pprint("Validating test test_execute_dq_policy")
        dq_rule = self.torch_client.get_policy(const.PolicyType.DATA_QUALITY, dq_policy_name)
        async_executor = dq_rule.execute(sync=False)
        if async_executor.errorMessage is None:
            async_execution_status = async_executor.get_status()
            pp.pprint('async_execution_status')
            pp.pprint(async_execution_status)
            async_execution_result = async_executor.get_result()
            pp.pprint('async_execution_result')
            pp.pprint(async_execution_result)

    def test_get_dq_policy_executions(self):
        pp.pprint("Validating test test_get_dq_policy_executions")
        dq_policy = self.torch_client.get_policy(identifier=dq_policy_name)
        dq_rule_executions = self.torch_client.policy_executions(dq_policy.id, RuleType.DATA_QUALITY)
        assert dq_rule_executions is not None

    # Reconciliation Policy Tests

    def test_get_recon_policy(self):
        pp.pprint("Validating test_get_recon_policy")
        recon_rule = self.torch_client.get_policy(const.PolicyType.RECONCILIATION, recon_policy_name)
        print(recon_rule)
        assert recon_rule is not None

    def test_get_recon_policy_without_type(self):
        pp.pprint("Validating test test_get_recon_policy_without_type")
        recon_rule = self.torch_client.get_policy(identifier=recon_policy_name)
        pp.pprint("recon_rule")
        pp.pprint(recon_rule)
        assert recon_rule is not None

    def test_get_all_recon_policy(self):
        pp.pprint("Validating test test_get_all_dq_policy")
        policy_filter = PolicyFilter(policyType=RuleType.RECONCILIATION, enable=True)
        recon_rules = self.torch_client.list_all_policies(filter=policy_filter)
        pp.pprint("recon_rules")
        pp.pprint(recon_rules)
        assert recon_rules is not None

    def test_cancel_recon_policy(self):
        pp.pprint("Validating test test_cancel_recon_policy")
        recon_rule = self.torch_client.get_policy(const.PolicyType.RECONCILIATION, recon_policy_name)
        pp.pprint("Validating test async_executor")
        async_executor = recon_rule.execute(sync=False)
        if async_executor.errorMessage is None:
            async_execution_status = async_executor.get_status()
            pp.pprint('async_execution_status')
            pp.pprint(async_execution_status)
            async_execution_cancel = async_executor.cancel()
            pp.pprint('async_execution_cancel')
            pp.pprint(async_execution_cancel)

    def test_execute_recon_policy(self):
        pp.pprint("Validating test test_execute_recon_policy")
        recon_rule = self.torch_client.get_policy(const.PolicyType.RECONCILIATION, recon_policy_name)
        async_executor = recon_rule.execute(sync=False)
        if async_executor.errorMessage is None:
            async_execution_status = async_executor.get_status()
            pp.pprint('async_execution_status')
            pp.pprint(async_execution_status)
            async_execution_result = async_executor.get_result()
            pp.pprint('async_execution_result')
            pp.pprint(async_execution_result)

    def test_get_recon_policy_executions(self):
        pp.pprint("Validating test test_get_recon_policy_executions")
        policy = self.torch_client.get_policy(identifier=recon_policy_name)
        recon_rule_executions = self.torch_client.policy_executions(policy.id, RuleType.RECONCILIATION)
        assert recon_rule_executions is not None

    # Data sources  tests
    def test_get_datasource(self):
        pp.pprint("Validating test test_get_datasource")
        datasource = self.torch_client.get_datasource(datasource_name, True)
        pp.pprint(datasource)
        assert datasource is not None

    def test_get_datasource_by_id(self):
        datasource = self.torch_client.get_datasource(datasource_name, True)
        pp.pprint("Validating test test_get_datasource_id")
        datasource_with_id = self.torch_client.get_datasource(datasource.id, False)
        pp.pprint(datasource_with_id)
        assert datasource_with_id is not None

    def test_get_all_data_sources(self):
        pp.pprint("Validating test test_get_all_data_sources")
        datasources = self.torch_client.get_datasources()
        pp.pprint(datasources)
        assert datasources is not None

    def test_get_ds_crawler_status(self):
        pp.pprint("Validating test test_get_ds_crawler_status")
        datasource = self.torch_client.get_datasource(datasource_name, True)
        status = datasource.get_crawler_status()
        pp.pprint(status)
        assert status is not None

    def test_start_crawler(self):
        pp.pprint("Validating test test_start_crawler")
        datasource = self.torch_client.get_datasource(datasource_name, False)
        start_crawler = datasource.start_crawler()
        pp.pprint(start_crawler)
        status = datasource.get_crawler_status()
        pp.pprint(status)
        assert status is not None

    # Assets tests
    def test_get_asset_by_uid(self):
        pp.pprint("Validating test test_get_asset_by_uid")
        asset = self.torch_client.get_asset(asset_uid_table)
        pp.pprint(asset)
        assert asset is not None

    def test_get_asset_by_id(self):
        pp.pprint("Validating test test_get_asset_by_id")
        asset = self.torch_client.get_asset(asset_uid_table)
        asset_with_id = self.torch_client.get_asset(asset.id)
        pp.pprint(asset_with_id)
        assert asset_with_id is not None

    def test_get_asset_metadata(self):
        pp.pprint("Validating test test_get_asset_metadata")
        asset = self.torch_client.get_asset(asset_uid_table)
        metadata_asset = asset.get_metadata()
        pp.pprint(metadata_asset)
        assert metadata_asset is not None

    def test_get_asset_sample_data_table(self):
        pp.pprint("Validating test test_get_asset_sample_data_table")
        asset = self.torch_client.get_asset(asset_uid_table)
        pp.pprint("Asset : {}".format(asset))
        sample_data_asset = asset.sample_data()
        pp.pprint(sample_data_asset)
        assert sample_data_asset is not None

    def test_get_asset_sample_data_file(self):
        pp.pprint("Validating test test_get_asset_sample_data_file")
        asset = self.torch_client.get_asset(asset_uid_file)
        sample_data_asset = asset.sample_data()
        pp.pprint(sample_data_asset)
        assert sample_data_asset is not None

    def test_get_asset_sample_data_kafka(self):
        pp.pprint("Validating test test_get_asset_sample_data_kafka")
        asset = self.torch_client.get_asset(asset_uid_kafka)
        sample_data_asset = asset.sample_data()
        pp.pprint(sample_data_asset)
        assert sample_data_asset is not None

    def test_get_asset_labels(self):
        pp.pprint("Validating test test_get_asset_labels")
        asset = self.torch_client.get_asset(asset_uid_table)
        labels_asset = asset.get_labels()
        pp.pprint(labels_asset)
        assert labels_asset is not None

    def test_add_asset_labels(self):
        pp.pprint("Validating test test_add_asset_labels")
        asset = self.torch_client.get_asset(asset_uid_table)
        asset.add_labels(labels=[AssetLabel('test12', 'shubh12'), AssetLabel('test22', 'shubh32')])
        labels_asset = asset.get_labels()
        pp.pprint(labels_asset)
        assert labels_asset is not None

    def test_add_asset_custom_metadata(self):
        pp.pprint("Validating test test_add_asset_custom_metadata")
        asset = self.torch_client.get_asset(asset_uid_table)
        asset.add_custom_metadata(
            custom_metadata=[CustomAssetMetadata('testcm1', 'shubhcm1'), CustomAssetMetadata('testcm2', 'shubhcm2')])
        metadata_asset = asset.get_metadata()
        pp.pprint(metadata_asset)
        assert metadata_asset is not None

    def test_profile_status(self):
        pp.pprint("Validating test test_profile_status")
        asset = self.torch_client.get_asset(asset_uid_table)
        latest_profile_status_asset = asset.get_latest_profile_status()
        pp.pprint(latest_profile_status_asset)
        assert latest_profile_status_asset is not None

    def test_cancel_profile(self):
        pp.pprint("Validating test test_cancel_profile")
        asset = self.torch_client.get_asset(asset_uid_table)
        latest_profile_status_asset = asset.get_latest_profile_status()
        pp.pprint("Latest profile status")
        pp.pprint(latest_profile_status_asset)
        if latest_profile_status_asset.status != 'IN PROGRESS':
            pp.pprint("Triggering profiling start request")
            start_profile_asset = asset.start_profile(ProfilingType.FULL)
            pp.pprint(start_profile_asset)
            time.sleep(15)
            profile_status = start_profile_asset.get_status()
            if profile_status['profileRequest']['status'] == 'IN PROGRESS':
                pp.pprint("Cancelling profiling request")
                cancel_res = start_profile_asset.cancel()
                pp.pprint(profile_status)
                assert cancel_res is not None
        else:
            pp.pprint("Profile execution request already in progress")

    def test_execute_profile(self):
        pp.pprint("Validating test test_execute_profile")
        asset = self.torch_client.get_asset(asset_uid_table)
        latest_profile_status_asset = asset.get_latest_profile_status()
        pp.pprint("Latest profile status")
        pp.pprint(latest_profile_status_asset)
        if latest_profile_status_asset.status != 'IN PROGRESS':
            start_profile_asset = asset.start_profile(ProfilingType.FULL)
            pp.pprint(start_profile_asset)
            profile_status = start_profile_asset.get_status()
            pp.pprint("Current profiling result status")
            pp.pprint(profile_status)
            assert profile_status is not None
        else:
            pp.pprint("Profile execution request already in progress")


if __name__ == '__main__':
    suite = unittest.TestSuite()

    # Pipeline Test cases
    suite.addTest(TestTorchClientIntegration('test_get_torch_version'))
    suite.addTest(TestTorchClientIntegration('test_get_supported_sdk_versions'))
    suite.addTest(TestTorchClientIntegration('test_create_pipeline'))
    suite.addTest(TestTorchClientIntegration('test_get_pipeline_by_uid'))
    suite.addTest(TestTorchClientIntegration('test_create_pipeline_run'))
    suite.addTest(TestTorchClientIntegration('test_latest_pipeline_run'))
    suite.addTest(TestTorchClientIntegration('test_create_root_span'))
    suite.addTest(TestTorchClientIntegration('test_create_job_for_pipeline'))
    suite.addTest(TestTorchClientIntegration('test_create_job_for_pipeline_run'))
    suite.addTest(TestTorchClientIntegration('test_create_child_span'))
    suite.addTest(TestTorchClientIntegration('test_update_pipeline_run'))
    suite.addTest(TestTorchClientIntegration('test_spans_pipeline_run'))
    suite.addTest(TestTorchClientIntegration('test_details_pipeline_run'))
    suite.addTest(TestTorchClientIntegration('test_get_pipelines'))
    suite.addTest(TestTorchClientIntegration('test_get_pipeline_runs'))
    suite.addTest(TestTorchClientIntegration('test_create_pipeline_run_with_continuation_id'))
    suite.addTest(TestTorchClientIntegration('test_delete_pipeline'))

    # DQ Policy Test Cases
    suite.addTest(TestTorchClientIntegration('test_execute_dq_policy'))
    suite.addTest(TestTorchClientIntegration('test_get_dq_policy'))
    suite.addTest(TestTorchClientIntegration('test_get_dq_policy_without_type'))
    suite.addTest(TestTorchClientIntegration('test_get_all_dq_policy'))
    suite.addTest(TestTorchClientIntegration('test_cancel_dq_policy'))
    suite.addTest(TestTorchClientIntegration('test_get_dq_policy_executions'))

    # Reconciliation Policy Test Cases
    suite.addTest(TestTorchClientIntegration('test_execute_recon_policy'))
    suite.addTest(TestTorchClientIntegration('test_get_recon_policy'))
    suite.addTest(TestTorchClientIntegration('test_get_recon_policy_without_type'))
    suite.addTest(TestTorchClientIntegration('test_get_all_recon_policy'))
    suite.addTest(TestTorchClientIntegration('test_cancel_recon_policy'))
    suite.addTest(TestTorchClientIntegration('test_get_recon_policy_executions'))

    # Data sources and Assets Test Cases
    suite.addTest(TestTorchClientIntegration('test_get_datasource'))
    suite.addTest(TestTorchClientIntegration('test_get_datasource_by_id'))
    suite.addTest(TestTorchClientIntegration('test_get_all_data_sources'))
    suite.addTest(TestTorchClientIntegration('test_get_ds_crawler_status'))
    suite.addTest(TestTorchClientIntegration('test_start_crawler'))
    suite.addTest(TestTorchClientIntegration('test_get_asset_by_uid'))
    suite.addTest(TestTorchClientIntegration('test_get_asset_by_id'))
    suite.addTest(TestTorchClientIntegration('test_get_asset_metadata'))
    suite.addTest(TestTorchClientIntegration('test_get_asset_sample_data_table'))
    suite.addTest(TestTorchClientIntegration('test_get_asset_sample_data_file'))
    suite.addTest(TestTorchClientIntegration('test_get_asset_sample_data_kafka'))
    suite.addTest(TestTorchClientIntegration('test_get_asset_labels'))
    suite.addTest(TestTorchClientIntegration('test_get_asset_labels'))
    suite.addTest(TestTorchClientIntegration('test_add_asset_labels'))
    suite.addTest(TestTorchClientIntegration('test_add_asset_custom_metadata'))
    suite.addTest(TestTorchClientIntegration('test_execute_profile'))
    suite.addTest(TestTorchClientIntegration('test_profile_status'))
    suite.addTest(TestTorchClientIntegration('test_cancel_profile'))

    unittest.TextTestRunner().run(suite)



from acceldata_sdk.errors import TorchSdkException
from acceldata_sdk.models.pipeline import CreatePipeline, Pipeline, PipelineRun
from acceldata_sdk.torch_http_client import TorchHttpClient
from acceldata_sdk.constants import RuleExecutionStatus, FailureStrategy, PolicyType, AssetSourceType, \
    MIN_TORCH_BACKEND_VERSION_SUPPORTED, SdkSupportedVersions, TorchBuildVersion

from acceldata_sdk.models.rule import RuleResource
from acceldata_sdk.models.common_types import PolicyExecutionRequest
from typing import List
import distutils
from semantic_version import Version, SimpleSpec
import logging
from acceldata_sdk.models.ruleExecutionResult import RuleResult, PolicyFilter, RuleType


class TorchClient:
    logger = logging.getLogger('torch')
    logger.setLevel(logging.INFO)

    """
            Description : Torch user client is used to send data to catalog server.
            :param url: (String) url of the catalog server
            :param timeout_ms: (Integer) timeout of the requests sending to catalog
            :param access_key: (String) Access key of API key. You can generate API key from torch UI's setting
            :param secret_key: (String) Secret key of API key.
            :param do_version_check: (bool) Enable/Disable version compatibility check between sdk and ADOC. By default version checks are disabled.
            Ex.  TorchClient = TorchUserClient(url='https://torch.acceldata.local:5443', access_key='OY2VVIN2N6LJ', secret_key='da6bDBimQfXSMsyyhlPVJJfk7Zc2gs')
    """

    def __init__(self, url, timeout_ms=10000, access_key: str = None, secret_key: str = None,
                 do_version_check: bool = False):
        """
                Description : Torch user client is used to send data to catalog server.
                :param url: (String) url of the catalog server
                :param timeout_ms: (Integer) timeout of the requests sending to catalog
                :param access_key: (String) Access key of API key. You can generate API key from torch UI's setting
                :param secret_key: (String) Secret key of API key.
                :param do_version_check: (bool) Enable/Disable version compatibility check between sdk and torch.
                Ex.  TorchClient = TorchUserClient(url='https://torch.acceldata.local:5443', access_key='OY2VVIN2N6LJ', secret_key='da6bDBimQfXSMsyyhlPVJJfk7Zc2gs')
        """
        if access_key is None and secret_key is None:
            raise Exception('Access key and secret key - required')
        self.client = TorchHttpClient(url=url, access_key=access_key, secret_key=secret_key, timeout_ms=timeout_ms)
        if isinstance(do_version_check, str):
            self.do_version_check = bool(distutils.util.strtobool(do_version_check))
        else:
            self.do_version_check = do_version_check
        if self.do_version_check:
            supported_versions = self.get_supported_sdk_versions()
            ver_comparator = SimpleSpec(f'>={supported_versions.minVersion},<={supported_versions.maxVersion}')
            if Version(MIN_TORCH_BACKEND_VERSION_SUPPORTED) not in ver_comparator:
                raise Exception(f'Torch supports sdk versions between {supported_versions.minVersion} and '
                                f'{supported_versions.maxVersion}')
        else:
            self.logger.info('Skipping version check')

    def get_supported_sdk_versions(self) -> SdkSupportedVersions:
        """
        Description:
            To get supported sdk versions
        :return: (SdkSupportedVersions) SdkSupportedVersions class instance
        """
        return self.client.get_supported_sdk_versions()

    def get_torch_version(self) -> TorchBuildVersion:
        """
        Description:
            To get supported torch version
        :return: (TorchBuildVersion)
        """
        return self.client.get_torch_version()

    def create_pipeline(self, pipeline: CreatePipeline) -> Pipeline:
        """
        Description:
            To create pipeline in torch catalog service
        :param pipeline: (CreatePipeline) class instance of the pipeline to be created
        :return: (Pipeline) newly created pipeline class instance
        """
        if pipeline.uid is None or pipeline.name is None:
            raise Exception('To create a pipeline, pipeline uid/name is required')
        return self.client.create_pipeline(pipeline)

    def get_pipeline(self, pipeline_identity) -> Pipeline:
        """
        Description:
            To get an existing pipeline from torch catalog
        :param pipeline_identity: uid or id of the pipeline
        :return:(Pipeline) pipeline class instance
        """
        return self.client.get_pipeline(pipeline_identity)

    def get_pipelines(self):
        """
        Description:
            To get an all pipelines from torch catalog
        :return:(List[PipelineInfo]) pipeline class instance
        """
        return self.client.get_pipelines()

    def get_spans(self, pipeline_run_id):
        """
        Description:
            To get an all spans from torch catalog
        :param pipeline_run_id: run_id of the pipeline
        :return:(List[Span]) pipeline class instance
        """
        return self.client.get_spans(pipeline_run_id)

    def get_pipeline_run(self, pipeline_run_id: str = None, continuation_id: str = None, pipeline_id: str = None) -> PipelineRun:
        """
        Description:
            To get an existing pipeline from torch catalog
        :param pipeline_run_id: run id of the pipeline run
        :param continuation_id: continuation id of the pipeline run
        :param pipeline_id: id of the pipeline. This is a mandatory parameter when run is being queried using continuation_id

        :return:(PipelineRun) pipeline run class instance
        """
        return self.client.get_pipeline_run(pipeline_run_id=pipeline_run_id, continuation_id=continuation_id, pipeline_id=pipeline_id)

    def get_pipeline_runs(self, pipeline_id) -> List[PipelineRun]:
        """
        Description:
            To get an all pipeline runs for a pipeline id from torch catalog
        :param pipeline_id: id of the pipeline
        :return:(List[PipelineRun]) List of pipeline run class instance
        """
        return self.client.get_pipeline_runs(pipeline_id)

    def get_datasource(self, assembly_identifier, properties=False):
        """
        Description:
            Find datasource by it's name or id in torch catalog
        :param assembly_identifier: name or id of the datasource given in torch
        :param properties: optional parameter, bool, to get datasource properties as well
        :return: (DataSource) datasource
        """
        return self.client.get_datasource(assembly_identifier, properties)

    def get_datasource_by_id(self, id: int, properties: bool = False):
        """
        Description:
            Find datasource by its id in torch catalog
        :param id: id of the datasource given in torch
        :return: (DataSource) datasource
        """
        return self.client.get_datasource_by_id(id, properties)

    def get_datasources(self, type: AssetSourceType = None):
        """
        Description:
            Find datasources by its type in torch catalog
        :param type: type of the datasource given in torch, optional
        :return: list(DataSource) datasource
        """
        return self.client.get_datasources(type)

    def get_all_datasources(self):
        """
        Description:
            list all datasources in torch catalog
        :return: (DataSource) list of datasource
        """
        return self.client.get_all_datasources()

    def start_crawler(self, datasource_name: str):
        """
        Description:
            Start crawler for datasource
        :param datasource_name: name of the datasource given in torch
        :return: (CrawlerStatus) CrawlerStatus
        """
        return self.client.start_crawler(datasource_name)

    def get_crawler_status(self, datasource_name: str):
        """
        Description:
            Get crawler status for datasource
        :param datasource_name: name of the datasource given in torch
        :return: (CrawlerStatus) CrawlerStatus
        """
        return self.client.get_crawler_status(datasource_name)

    def get_asset_types(self):
        """
        Description:
            get all asset types supported in torch catalog
        :return: list of asset types
        """
        return self.client.get_all_asset_types()

    def get_profile_status(self, asset_id: int, req_id: int):
        """
        Description:
            get status of asset profile for given request id
        :return: profiling status
        """
        return self.client.get_profile_request_details(asset_id=asset_id, req_id=req_id)

    def get_all_source_types(self):
        """
        Description:
            get all source types supported in torch catalog
        :return: list of all source type
        """
        return self.client.get_all_source_types()

    def get_property_templates(self):
        pass

    def get_connection_types(self):
        """
        Description:
            get all connection types supported in torch catalog
        :return: list of all connection types
        """
        return self.client.get_connection_types()

    def get_tags(self):
        return self.client.get_tags()

    def get_analysis_pipeline(self, id: int):
        return self.client.get_analysis_pipeline(id)

    def get_asset(self, identifier):
        """"
            Description:
                Find an asset of the datasource
            :param identifier: uid or ID of the asset
        """
        return self.client.get_asset(identifier)

    def get_policy(self, type: PolicyType = None, identifier=None) -> RuleResource:
        """"
            Description:
                Find policy
            :param identifier: ID or name of the policy
            :param type: (Optional) type of the policy
        """
        if identifier is None:
            raise ValueError("Parameter 'identifier' is mandatory.")

        if type is None:
            return self.client.get_policy(identifier)

        if type.value == PolicyType.DATA_QUALITY.value:
            return self.client.get_dq_rule(identifier)
        elif type.value == PolicyType.RECONCILIATION.value:
            return self.client.get_reconciliation_rule(identifier)
        else:
            raise TorchSdkException('Invalid policy type passed.')


    def list_all_policies(self, filter: PolicyFilter, page=0, size=25, withLatestExecution=True, sortBy='updatedAt:DESC'):
        """
        Description:
            To list all policies based on filter
        :param filter: (Enum PolicyFilter) Type of policies to be executed.
        PolicyFilter can have the following parameters
        period: Time period for which list should be filtered. It is an enum ExecutionPeriod
        tags: tags on which list should be filtered
        lastExecutionResult: lastExecution result on which list should be filtered
        asset: Assets for which policies should be filtered. This list should be of objects of type Asset returned by get_asset
        data_source:  Datasource for which policies should be filtered. This list should be of objects of type Datasource
        returned by get_datasource
        policyType: Type of policy policies should be filtered. This is an enum of type RuleType
        enable: To filter only for enabled policies set enable to True
        active: : To filter only for active policies set active to True
        :param page: (Int) page number of query output
        :param size: (Int) Size of each page
        :param sortBy: (Str) sorting order
        """
        return self.client.get_all_rules(filter=filter, page=page, size=size, sortBy=sortBy,
                                         withLatestExecution=withLatestExecution)

    def create_connection(self, create_connection):
        return self.client.create_connection(create_connection)

    def check_connection(self, connection):
        return self.client.check_connection(connection)

    def execute_dq_rule(self, rule_id, incremental=False, pipeline_run_id=None,
                        policy_execution_request: PolicyExecutionRequest = None):
        self.logger.info(
            "Invoked execute_dq_rule of the torch_client\n"
            f"Incremental flag: {incremental}\n"
            f"Policy Execution Request: {policy_execution_request.to_dict() if policy_execution_request else 'None'}"
        )

        return self.client.execute_dq_rule(rule_id=rule_id, incremental=incremental,
                                           policy_execution_request=policy_execution_request)

    def execute_reconciliation_rule(self, rule_id, incremental=False,
                                    policy_execution_request: PolicyExecutionRequest = None):
        self.logger.info(
            "Invoked execute_reconciliation_rule of the torch_client\n"
            f"Incremental flag: {incremental}\n"
            f"Policy Execution Request: {policy_execution_request.to_dict() if policy_execution_request else 'None'}"
        )

        return self.client.execute_reconciliation_rule(rule_id=rule_id, incremental=incremental, pipeline_run_id=None,
                                                       policy_execution_request=policy_execution_request)

    def get_dq_rule_execution_details(self, execution_id):
        return self.client.get_dq_rule_execution_details(execution_id)

    def get_reconciliation_rule_execution_details(self, execution_id):
        return self.client.get_reconciliation_rule_execution_details(execution_id)

    def cancel_rule_execution(self, execution_id):
        return self.client.cancel_rule_execution(execution_id)

    def enable_rule(self, rule_id):
        return self.client.enable_rule(rule_id)

    def disable_rule(self, rule_id):
        return self.client.disable_rule(rule_id)

    def get_reconciliation_rule_result(self, execution_id) -> RuleResult:
        return self.client.get_reconciliation_rule_result(execution_id)

    def get_dq_rule_result(self, execution_id) -> RuleResult:
        return self.client.get_dq_rule_result(execution_id)

    def execute_policy(self, policy_type: PolicyType, policy_id, sync=True, incremental=False,
                       failure_strategy: FailureStrategy = FailureStrategy.DoNotFail, pipeline_run_id=None,
                       policy_execution_request: PolicyExecutionRequest = None):
        """
        Description:
            To execute policies synchronously and asynchronously
        :param policy_type: (PolicyType) Type of rule to be executed
        :param policy_id: (String) id of the rule to be executed
        :param sync: (bool) optional Set it to False if asynchronous execution has to be done
        :param incremental: (bool) optional Set it to True if full execution has to be done
        :param failure_strategy: (enum) optional Set it to decide if it should fail at error,
            fail at warning or never fail
        :param pipeline_run_id: (long) optional Run id of the pipeline run where the policy is being executed. This can
            be used to link a policy execution id with a particular pipeline run id
        :param policy_execution_request: (PolicyExecutionRequest) An optional parameter that allows you to provide
        additional options for executing the policy. It is an instance of the PolicyExecutionRequest modules class,
        which contains various properties that can be used to customize the policy execution, such as `executionType`,
        `markerConfigs`, `ruleItemSelections`, and more.
        """
        self.logger.info(f"Policy Execution Request: {policy_execution_request.to_dict() if policy_execution_request else 'None'}")

        return self.client.execute_rule(policy_type, policy_id, sync, incremental, failure_strategy, pipeline_run_id,
                                        policy_execution_request)

    def get_policy_status(self, policy_type, execution_id) -> RuleExecutionStatus:
        """
        Description:
            To get status of rule execution
        :param policy_type: (PolicyType) Type of rule to be executed
        :param execution_id: (String) ID of execution of the rule previously executed.
        """
        return self.client.get_rule_status(policy_type, execution_id)

    def get_policy_execution_result(self, policy_type: PolicyType, execution_id,
                                    failure_strategy: FailureStrategy = FailureStrategy.DoNotFail):
        """
        Description:
            To get result of rule execution
        :param policy_type: (PolicyType) Type of rule to be executed
        :param execution_id: (String) ID of execution of the rule previously executed.
        :param failure_strategy: (enum) optional Set it to decide if it should fail at error,
            fail at warning or never fail
        """
        return self.client.get_rule_execution_result(policy_type, execution_id, failure_strategy)

    def policy_executions(self, identifier, policy_type: RuleType, page=0, size=25, sortBy='finishedAt:DESC'):
        """
        Description:
            To list rule executions
        :param identifier: id or name of executed rule
        :param policy_type: (Enum RuleType) Type of executed rule
        :param page: (Int) page number of query output
        :param size: (Int) Size of each page
        :param sortBy: (Str) sorting order
        """
        return self.client.policy_executions(identifier, policy_type, page, size, sortBy)
